# pylint: disable = redefined-outer-name

import os
import pytest

from universum.lib.gravity import construct_component
from universum.modules.vcs.perforce_vcs import PerforceMainVcs
from . import utils
from .perforce_utils import PerforceWorkspace


class DiffParameters:
    def __init__(self, perforce_workspace: PerforceWorkspace):
        self.perforce_workspace: PerforceWorkspace = perforce_workspace

        for env_var in ["P4_PATH", "P4_MAPPINGS"]:
            try:
                del os.environ[env_var]
            except KeyError:
                pass

        # TODO: move this block to utils
        settings = utils.create_empty_settings("main")
        settings.PerforceVcs.port = perforce_workspace.server.port
        settings.PerforceVcs.user = perforce_workspace.server.user
        settings.PerforceVcs.password = perforce_workspace.server.password
        settings.Output.type = "term"
        settings.ProjectDirectory.project_root = str(self.perforce_workspace.root_directory) + "/../new_workspace"
        settings.PerforceWithMappings.mappings = [self.perforce_workspace.depot + " /..."]
        settings.PerforceMainVcs.force_clean = True
        settings.PerforceMainVcs.client = "new_client"

        self.perforce: PerforceMainVcs = construct_component(PerforceMainVcs, settings)

        self.perforce.connect()
        self.perforce.create_workspace()

    def run_diff(self, shelve):
        if not self.perforce.p4.run_opened():
            self.perforce.shelve_cls = [shelve["change"]]
            self.perforce.unshelve()
            files_to_compare = self.perforce.copy_cl_files_and_revert()
            is_opened = self.perforce.p4.run_opened()
            return files_to_compare, is_opened
        return None, "There is some opened files in new_workspace. Please, revert them before running tests."

    def exit(self):
        self.perforce.finalize()


@pytest.fixture()
def diff_parameters(perforce_workspace: PerforceWorkspace):
    params = None
    try:
        params = DiffParameters(perforce_workspace)
        yield params
    finally:
        if params:
            params.exit()


def test_p4_c_and_revert(diff_parameters):  # pylint: disable = too-many-locals
    p4 = diff_parameters.perforce_workspace.p4
    test_dir = diff_parameters.perforce_workspace.root_directory / "test_files"
    test_dir.mkdir()

    def create_file(filename):
        cur_file = test_dir / filename
        p4.run("add", str(cur_file))
        p4.run("edit", str(cur_file))
        cur_file.write_text(f"import os\n\nprint('File {0} has no special modifiers.')\n"
                       f"print(os.name)\nprint(os.getcwd())\nprint(os.strerror(3))\n"
                       f"print('File change type: {filename}')\n")

    create_file("edit")
    create_file("move")
    create_file("rename")
    create_file("integrate")
    create_file("integrated")
    create_file("branch")
    create_file("delete")

    change = p4.run_change("-o")[0]
    change["Description"] = "Submit test files"
    p4.run_submit(change)

    # open for edit for edit, move/add and move/rename files
    for cur_file in [test_dir / "edit", test_dir / "move", test_dir / "rename"]:
        p4.run("edit", str(cur_file))

    # edit
    (test_dir / "edit").write_text(f"import os\n\nprint('File {test_dir / 'edit'} has no special modifiers.')\n"
                                f"print(os.name)\nprint(os.getegid())\nprint(os.ctermid())\n"
                                f"print('File change type: \"edit\"')\n")

    # move, rename
    p4.run("move", test_dir / "move", test_dir / "moved/move")
    p4.run("rename", test_dir / "rename", test_dir / "renamed_rename")
    # integrate
    p4.run("integrate", test_dir / "integrate", test_dir / "integrated")
    p4.run("resolve", "-at")
    # branch
    p4.run("integrate", test_dir / "branch", test_dir / "moved/branch_to")
    # delete
    p4.run("delete", test_dir / "delete")
    # add
    add = test_dir / "add"
    add.write_text("\nprint('new file')\nprint('not in repo, only for shelve.')")
    p4.run("add", add)

    # make change
    change = p4.fetch_change()
    change["Description"] = "Test revert #1"
    saved_change = p4.save_change(change)[0].split()[1]

    shelve = p4.run_shelve("-fc", saved_change)[0]

    diff, is_files_opened = diff_parameters.run_diff(shelve)
    assert not is_files_opened

    new_test_dir = str(test_dir).replace("workspace", "new_workspace")
    new_temp = new_test_dir.replace("new_workspace", "new_workspace/new_temp")
    expected_path = [
        ("test_files/add", new_temp + "/add", None),
        (None, None, new_test_dir + '/delete'),
        ("test_files/edit", new_temp + "/edit", new_test_dir + "/edit"),
        ("test_files/integrated", new_temp + "/integrated", new_test_dir + "/integrated"),
        ("test_files/moved/branch_to", new_temp + "/moved/branch_to", None),
        ("test_files/moved/move", new_temp + "/moved/move", new_test_dir + "/move"),
        ("test_files/renamed_rename", new_temp + "/renamed_rename", new_test_dir + "/rename")
    ]

    for result, expected in zip(diff, expected_path):
        assert result == expected



from __future__ import annotations

import contextlib
from abc import ABC, abstractmethod
from typing import Callable, ClassVar, List, Optional, TypeVar, Generator

from typing_extensions import TypedDict

from .output import HasOutput
from ..configuration_support import Step, Configuration
from ..lib.ci_exception import SilentAbortException, CriticalCiException
from ..lib.gravity import Dependency, Module

__all__ = [
    "HasStructure",
    "Block",
    "RunningStepBase"
]


class Block:
    """
    >>> b1 = Block('Build for all supported platforms')
    >>> str(b1)
    ' Build for all supported platforms - Success'

    >>> b2 = Block('Build Android', b1)
    >>> str(b2)
    '1. Build Android - Success'

    >>> b3 = Block('Build Tizen', b1)
    >>> str(b3)
    '2. Build Tizen - Success'

    >>> b3.status = 'Failed'
    >>> str(b3)
    '2. Build Tizen - Failed'
    >>> b3.is_successful()
    False

    >>> b4 = Block('Run tests', b2)
    >>> str(b4)
    '1.1. Run tests - Success'
    >>> b4.is_successful()
    True
    >>> b2 is b4.parent
    True
    """

    def __init__(self, name: str, parent: Optional[Block] = None) -> None:
        self.name: str = name
        self.status: str = "Success"
        self.children: List[Block] = []

        self.parent: Optional[Block] = parent
        self.number: str = ''
        if parent:
            parent.children.append(self)
            self.number = f"{parent.number}{len(parent.children)}."

    def __str__(self) -> str:
        result = self.number + ' ' + self.name
        return f"{result} - {self.status}" if not self.children else result

    def is_successful(self) -> bool:
        return self.status == "Success"


class BackgroundStepInfo(TypedDict):
    name: str
    block: Block
    process: RunningStepBase
    is_critical: bool
    has_artifacts: bool


class RunningStepBase(ABC):

    @abstractmethod
    def start(self) -> None:
        pass

    @abstractmethod
    def finalize(self) -> None:
        pass

    @abstractmethod
    def get_error(self) -> Optional[str]:
        pass

    @abstractmethod
    def collect_artifacts(self) -> None:
        pass


class StructureHandler(HasOutput):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.current_block: Optional[Block] = Block("Universum")
        self.configs_current_number: int = 0
        self.configs_total_count: int = 0
        self.active_background_steps: List[BackgroundStepInfo] = []
        self.step_num_len: int = 0
        self.group_numbering: str = ""

    def open_block(self, name: str) -> None:
        new_block = Block(name, self.current_block)
        self.current_block = new_block

        self.out.open_block(new_block.number, name)

    def close_block(self) -> None:
        if self.current_block is not None:
            block: Block = self.current_block
            self.current_block = self.current_block.parent
            self.out.close_block(block.number, block.name, block.status)

    def log_critical_block_failure(self) -> None:
        self.out.log_skipped("Critical step failed. All further configurations will be skipped")

    def log_skipped_block(self, name):
        new_skipped_block = Block(name, self.current_block)
        new_skipped_block.status = "Skipped"

        self.out.log_skipped(new_skipped_block.number + name + " skipped because of critical step failure")

    def fail_current_block(self, error: str = ""):
        block: Block = self.get_current_block()
        self.fail_block(block, error)

    def fail_block(self, block, error: str = ""):
        if error:
            self.out.log_error(error)
        block.status = "Failed"
        self.out.report_build_problem(block.name + " " + block.status)

    def get_current_block(self):
        return self.current_block

    # The exact block will be reported as failed only if pass_errors is False
    # Otherwise the exception will be passed to the higher level function and handled there
    T = TypeVar('T')

    @contextlib.contextmanager
    def block(self, *, block_name: str, pass_errors: bool) -> Generator:
        self.open_block(block_name)
        try:
            yield
        except SilentAbortException:
            raise
        except CriticalCiException as e:  # system/environment step failed
            self.fail_current_block(str(e))
            raise SilentAbortException() from e
        except Exception as e:  # unexpected failure, should not occur
            if pass_errors is True:
                raise
            self.fail_current_block(str(e))
        finally:
            self.close_block()

    def execute_one_step(self, configuration: Step,
                         step_executor: Callable[[Step], RunningStepBase]) -> RunningStepBase:
        process: RunningStepBase = step_executor(configuration)
        process.start()
        if process.get_error() is not None:
            return process
        if not configuration.background:
            process.finalize()
            return process
        self.out.log("This step is marked to be executed in background")
        has_artifacts: bool = bool(configuration.artifacts) or bool(configuration.report_artifacts)
        self.active_background_steps.append({'name': configuration.name,
                                             'block': self.get_current_block(),
                                             'process': process,
                                             'is_critical': configuration.critical,
                                             'has_artifacts': has_artifacts})
        return process

    def finalize_background_step(self, background_step: BackgroundStepInfo) -> bool:
        process: RunningStepBase = background_step['process']
        process.finalize()
        error: Optional[str] = process.get_error()
        if error is not None:
            self.fail_block(background_step['block'], error)
            self.fail_current_block()
            return False

        return True

    def process_one_step(self, merged_item: Step, step_executor: Callable, skip_execution: bool) -> bool:
        """
        Process one step: either execute it or skip if the skip_execution flag is set.
        :param merged_item: Step to execute.
        :param step_executor: Function that executes one step.
        :param skip_execution: If True, the step will be skipped, but reported to the output.
        :return: True if step was successfully executed, False otherwise. Skipping is considered success.
        """
        self.configs_current_number += 1
        numbering: str = f" [ {self.configs_current_number:>{self.step_num_len}}/{self.configs_total_count} ] "
        step_label: str = numbering + merged_item.name

        if skip_execution:
            self.log_skipped_block(numbering + "'" + merged_item.name + "'")
            return True

        process: Optional[RunningStepBase] = None
        error: Optional[str] = None
        # Here pass_errors=False, because any exception while executing build step
        # can be step-related and may not affect other steps
        with self.block(block_name=step_label, pass_errors=False):
            process = self.execute_one_step(merged_item, step_executor)
            error = process.get_error()
            if error and not merged_item.is_conditional:
                self.fail_current_block(error)
        has_artifacts: bool = bool(merged_item.artifacts) or bool(merged_item.report_artifacts)
        if not merged_item.background and has_artifacts:
            with self.block(block_name=f"Collecting artifacts for the '{merged_item.name}' step", pass_errors=False):
                process.collect_artifacts()

        return error is None

    def execute_steps_recursively(self, parent: Step,
                                  children: Configuration,
                                  step_executor: Callable,
                                  skip_execution: bool) -> bool:
        """
        :param parent: Parent step. Used for calculating actual attributes of child steps.
        :param children: Configuration object containing child steps.
        :param step_executor: Function that executes one step.
        :param skip_execution: If True, all steps will be skipped, but reported to the output.
        :return: True if all child steps were successful, False otherwise.
        """
        # step_executor is [[Step], Step], but referring to Step creates circular dependency

        some_step_failed: bool = False
        for child in children.configs:
            merged_item: Step = parent + child
            current_step_failed: bool
            if child.children:
                step_label: str = self.group_numbering + merged_item.name

                # Here pass_errors=True, because any exception outside executing build step
                # is not step-related and should stop script executing
                with self.block(block_name=step_label, pass_errors=True):
                    current_step_failed = not self.execute_steps_recursively(merged_item, child.children, step_executor,
                                                                             skip_execution)
            elif child.is_conditional:
                conditional_step_succeeded: bool = self.process_one_step(merged_item, step_executor,
                                                                         skip_execution=False)
                step_to_execute: Optional[Configuration] = merged_item.if_succeeded if conditional_step_succeeded \
                    else merged_item.if_failed
                if step_to_execute:  # branch step can be not set
                    self.execute_steps_recursively(parent=Step(),
                                                   children=step_to_execute,
                                                   step_executor=step_executor,
                                                   skip_execution=False)
                current_step_failed = False  # conditional step should be always successful
            else:
                if merged_item.finish_background and self.active_background_steps:
                    self.out.log("All ongoing background steps should be finished before next step execution")
                    if not self.report_background_steps():
                        skip_execution = True
                current_step_failed = not self.process_one_step(merged_item, step_executor, skip_execution)

            if current_step_failed:
                some_step_failed = True

                if child.critical:
                    self.log_critical_block_failure()
                    skip_execution = True

        if some_step_failed:
            self.fail_current_block()

        return not some_step_failed

    def report_background_steps(self) -> bool:
        result: bool = True
        for item in self.active_background_steps:
            with self.block(block_name="Waiting for background step '" + item['name'] + "' to finish...",
                            pass_errors=True):
                if not self.finalize_background_step(item) and item['is_critical']:
                    result = False
                    self.out.log_skipped("The background step '" + item['name'] + "' failed, and as it is critical, "
                                                                                  "all further steps will be skipped")
            if item['has_artifacts']:
                with self.block(block_name=f"Collecting artifacts for the '{item['name']}' step", pass_errors=False):
                    item['process'].collect_artifacts()

        self.out.log("All ongoing background steps completed")
        self.active_background_steps = []
        return result

    def execute_step_structure(self, configs: Configuration, step_executor) -> None:
        for config in configs.all():
            self.configs_total_count += 1
            if config.is_conditional:
                self.configs_total_count += 1
        self.step_num_len = len(str(self.configs_total_count))
        self.group_numbering = f" [ {'':{self.step_num_len}}+{'':{self.step_num_len}} ] "

        self.execute_steps_recursively(Step(), configs, step_executor, False)

        if self.active_background_steps:
            with self.block(block_name="Reporting background steps", pass_errors=False):
                self.report_background_steps()

    def _build_step_name(self, name):
        step_num_len = len(str(self.configs_total_count))
        numbering = f" [ {self.configs_current_number:>{step_num_len}}/{self.configs_total_count} ] "
        return numbering + name


class HasStructure(Module):
    structure_factory: ClassVar = Dependency(StructureHandler)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.structure: StructureHandler = self.structure_factory()



# pylint: disable = redefined-outer-name, abstract-method

import pathlib
import pytest

from universum.modules.vcs.github_app_vcs import GithubToken
from . import utils
from .git_utils import GitClient


class ReportEnvironment(utils.BaseTestEnvironment):
    def __init__(self, client: GitClient, directory: pathlib.Path):
        super().__init__(client, directory, "main", "")

        self.settings.Vcs.type = "ghapp"
        self.settings.MainVcs.report_to_review = True
        self.settings.GitVcs.repo = client.server.url
        commit_id = str(client.repo.remotes.origin.refs[client.server.target_branch].commit)
        self.settings.GitMainVcs.checkout_id = commit_id
        self.settings.GithubToken.integration_id = "1234"
        self.settings.GithubToken.key = "this is key"
        self.settings.GithubTokenWithInstallation.installation_id = "5678"
        self.settings.GithubAppMainVcs.check_id = "123"
        self.settings.GithubAppMainVcs.api_url = "http://localhost/"
        self.settings.Reporter.report_start = True
        self.settings.Reporter.report_success = True

        repo_name = str(client.root_directory).rsplit("client", 1)[0]
        self.path: str = "http://localhost/repos" + repo_name + "server/check-runs/123"


@pytest.fixture()
def report_environment(git_client: GitClient, tmp_path: pathlib.Path):
    yield ReportEnvironment(git_client, tmp_path)


def test_github_run(report_environment: ReportEnvironment, monkeypatch):
    monkeypatch.setattr(GithubToken, 'get_token', lambda *args, **kwargs: "this is token")

    collected_http = report_environment.run_with_http_server(url=report_environment.path, method="PATCH")
    collected_http.assert_request_body_contained("status", "in_progress")
    collected_http.assert_request_body_contained("status", "completed")
    collected_http.assert_request_body_contained("conclusion", "success")



__all__ = [
    "CiException",
    "CriticalCiException",
    "SilentAbortException"
]


class CiException(Exception):
    pass


class CriticalCiException(Exception):
    pass


class SilentAbortException(Exception):
    def __init__(self, application_exit_code: int = 1) -> None:
        super().__init__()
        self.application_exit_code = application_exit_code



# pylint: disable = redefined-outer-name, invalid-name

from argparse import ArgumentError
from unittest import mock
import pytest

from universum.lib.gravity import construct_component, define_arguments_recursive, Dependency
from universum.lib.gravity import get_dependencies
import universum.lib.module_arguments
import universum.lib.gravity


@pytest.fixture()
def mock_module():
    class MockedModule:
        def __new__(cls, *args, **kwargs):
            return old_module.__new__(cls, *args, **kwargs)

    old_module = universum.lib.gravity.Module
    universum.lib.gravity.Module = MockedModule

    yield MockedModule

    universum.lib.gravity.Module = old_module


def mock_define_arguments(klass):
    patcher = mock.patch.object(klass, "define_arguments", create=True)
    patcher.start()
    return klass


def test_dependencies_simple(mock_module):
    class OnlyDep(mock_module):
        pass

    class RootOneUser(mock_module):
        dep = Dependency(OnlyDep)

    assert get_dependencies(RootOneUser) == [RootOneUser, OnlyDep]
    assert get_dependencies(OnlyDep) == [OnlyDep]


def test_dependencies_two_users(mock_module):
    class Common(mock_module):
        pass

    class FirstDep(mock_module):
        dep = Dependency(Common)

    class SecondDep(mock_module):
        dep = Dependency(Common)

    class RootTwoUsers(mock_module):
        first = Dependency(FirstDep)
        second = Dependency(SecondDep)

    dependencies = get_dependencies(RootTwoUsers)
    for x in (RootTwoUsers, FirstDep, SecondDep, Common):
        assert x in dependencies


def test_define_arguments_mock_no_call(mock_module):
    @mock_define_arguments
    class FirstDependency(mock_module):
        pass

    @mock_define_arguments
    class NotADependency(mock_module):
        pass

    class RootDefineArgs(mock_module):
        dep = Dependency(FirstDependency)

    FirstDependency.define_arguments.assert_not_called()
    NotADependency.define_arguments.assert_not_called()

    parser = mock.MagicMock()
    define_arguments_recursive(RootDefineArgs, parser)

    assert parser.dest_prefix == ""
    FirstDependency.define_arguments.assert_called_with(parser)
    NotADependency.define_arguments.assert_not_called()


def test_define_arguments_3(mock_module):
    class ChainEnd(mock_module):
        @staticmethod
        def define_arguments(parser):
            group_parser = parser.get_or_create_group("ChainEnd group")
            group_parser.add_argument('--option3')

    class ChainSecond(mock_module):
        dep = Dependency(ChainEnd)

        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--option2')

    class ChainStart(mock_module):
        dep = Dependency(ChainSecond)

        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--option1')

    parser = universum.lib.module_arguments.ModuleArgumentParser()
    define_arguments_recursive(ChainStart, parser)
    settings = parser.parse_args(["--option2=abc"])
    assert settings.ChainStart.option1 is None
    assert settings.ChainSecond.option2 == "abc"
    assert settings.ChainEnd.option3 is None

    parser = universum.lib.module_arguments.ModuleArgumentParser()
    define_arguments_recursive(ChainSecond, parser)
    settings = parser.parse_args(["--option2=def"])
    assert 'ChainStart' not in dir(settings)
    assert settings.ChainSecond.option2 == "def"
    assert settings.ChainEnd.option3 is None

    parser = universum.lib.module_arguments.ModuleArgumentParser()
    define_arguments_recursive(ChainEnd, parser)
    settings = parser.parse_args([])
    assert 'ChainStart' not in dir(settings)
    assert 'ChainSecond' not in dir(settings)
    assert settings.ChainEnd.option3 is None


def test_define_arguments_inheritance(mock_module, capsys):
    class ParentModule(mock_module):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--parent')

    class InheritedModule1(ParentModule):
        pass

    class InheritedModule2(ParentModule):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--child')

    class WrongInheritedModule(ParentModule):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--parent')

    parser = universum.lib.module_arguments.ModuleArgumentParser()
    define_arguments_recursive(ParentModule, parser)
    settings = parser.parse_args(["--parent=abc"])
    assert settings.ParentModule.parent == "abc"
    with pytest.raises(SystemExit) as exception_info:
        parser.parse_args(["--child=def"])
    assert exception_info.value.code == 2
    assert "unrecognized arguments: --child=def" in capsys.readouterr()[1]

    parser = universum.lib.module_arguments.ModuleArgumentParser()
    define_arguments_recursive(InheritedModule1, parser)
    settings = parser.parse_args(["--parent=abc"])
    assert settings.ParentModule.parent == "abc"

    parser = universum.lib.module_arguments.ModuleArgumentParser()
    define_arguments_recursive(InheritedModule2, parser)
    settings = parser.parse_args(["--parent=abc", "--child=def"])
    assert settings.ParentModule.parent == "abc"
    assert settings.InheritedModule2.child == "def"

    parser = universum.lib.module_arguments.ModuleArgumentParser()
    with pytest.raises(ArgumentError) as exception_info:
        define_arguments_recursive(WrongInheritedModule, parser)
    assert "conflicting option string: --parent" in str(exception_info.value)


def test_settings_access(mock_module):
    class S(mock_module):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--option')

    class R(mock_module):
        dep = Dependency(S)

        def __init__(self):
            self.s = self.dep()

    settings = universum.lib.module_arguments.ModuleNamespace()
    child_settings = universum.lib.module_arguments.ModuleNamespace()
    child_settings.option = "abc"
    setattr(settings, 'S', child_settings)
    r = construct_component(R, settings)

    # get existing settings
    assert r.s.settings.option == "abc"

    # get non-existing settings
    with pytest.raises(AttributeError) as exception_info:
        print(r.settings.option)
    assert "'R' object has no setting 'option'" in str(exception_info.value)

    # get non-existing option
    with pytest.raises(AttributeError) as exception_info:
        print(r.s.settings.another_option)
    assert "'S' object has no setting 'another_option'" in str(exception_info.value)

    # set non-existing option
    with pytest.raises(AttributeError) as exception_info:
        r.s.settings.another_option = "def"
    assert "'S' object has no setting 'another_option'" in str(exception_info.value)

    # set non-existing settings
    with pytest.raises(AttributeError) as exception_info:
        r.settings.option = "hij"
    assert "'R' object has no setting 'option'" in str(exception_info.value)


def make_settings(name, value):
    settings = universum.lib.module_arguments.ModuleNamespace()
    child_settings = universum.lib.module_arguments.ModuleNamespace()
    child_settings.option = value

    if not isinstance(name, str):
        name = name.__name__
    setattr(settings, name, child_settings)
    return settings


def test_construct_component(mock_module):
    class Q(mock_module):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--option')

        def __init__(self):
            self.member = self.settings.option

        def empty_init(self):
            pass

    class P(mock_module):
        dep = Dependency(Q)

        def __init__(self):
            self.q = 123

        def init_with_dependency(self):
            self.q = self.dep()

    # no reference to Q
    p1 = construct_component(P, make_settings(Q, ""))
    assert p1.q == 123

    # create dynamic reference to Q by calling method
    p2 = construct_component(P, make_settings(Q, "def"))
    p2.init_with_dependency()
    assert p2.q.member == "def"

    # create static reference to Q by rewriting __init__
    P.__init__ = P.init_with_dependency
    p3 = construct_component(P, make_settings(Q, "ghi"))
    assert p3.q.member == "ghi"

    # remove __init__ of Q to make sure "member" is present because of __init__ call
    Q.__init__ = Q.empty_init
    p4 = construct_component(P, make_settings(Q, "jkl"))
    assert 'member' not in dir(p4.q)


def test_construct_component_same_instance(mock_module):
    class V(mock_module):
        pass

    class U(mock_module):
        dep = Dependency(V)

        def __init__(self):
            self.v = self.dep()

    class T(mock_module):
        dep1 = Dependency(U)
        dep2 = Dependency(V)

        def __init__(self):
            self.u = self.dep1()
            self.v = self.dep2()

    settings = universum.lib.module_arguments.ModuleNamespace()
    t = construct_component(T, settings)
    assert t.v == t.u.v
    assert "T" in repr(t)
    assert "U" in repr(t.u)
    assert "V" in repr(t.v)

    u = construct_component(U, settings)
    assert t.u == u

    v = construct_component(V, settings)
    assert t.v == v


def test_construct_component_same_name(mock_module):

    def get_class(parameter):
        class TheOnlyClass(mock_module):
            the_only_parameter = parameter

        return TheOnlyClass

    settings = universum.lib.module_arguments.ModuleNamespace()
    first_object = construct_component(get_class("abc"), settings)
    second_object = construct_component(get_class("def"), settings)

    assert first_object.the_only_parameter == "abc"
    assert second_object.the_only_parameter == "def"


def test_additional_init_parameters(mock_module):
    class W(mock_module):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--option')

        def __init__(self, param, named_parm):
            self.member1 = self.settings.option
            self.member2 = param
            self.member3 = named_parm

    class X(mock_module):
        dep = Dependency(W)

        def __init__(self):
            self.w = self.dep(123, named_parm=456)

    x = construct_component(X, make_settings(W, "abc"))
    assert x.w.member1 == "abc"
    assert x.w.member2 == 123
    assert x.w.member3 == 456


def parse_settings(klass, arguments):
    parser = universum.lib.module_arguments.ModuleArgumentParser()
    define_arguments_recursive(klass, parser)
    return parser.parse_args(arguments)


def test_construct_component_inheritance(mock_module):
    class BaseModule(mock_module):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--base_option')

        def __init__(self, base_param, **kwargs):
            super().__init__(**kwargs)
            self.base_option = self.settings.base_option
            self.base_param = base_param

    class DerivedModule(BaseModule):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--derived_option')

        def __init__(self, **kwargs):
            super().__init__(456, **kwargs)
            self.derived_option = self.settings.derived_option
            self.derived_member = 123

    settings = parse_settings(DerivedModule, ["--base_option=abc", "--derived_option=def"])
    assert settings.BaseModule.base_option == "abc"
    assert settings.DerivedModule.derived_option == "def"
    derived = construct_component(DerivedModule, settings)
    assert derived.base_option == "abc"
    assert derived.derived_option == "def"
    assert derived.derived_member == 123
    assert derived.base_param == 456


def test_construct_component_multiple_inheritance(mock_module):
    class Base1(mock_module):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--base1_option')

        def __init__(self, base1_param, **kwargs):
            super().__init__(**kwargs)
            self.base1_option = self.settings.base1_option
            self.base1_param = base1_param

    class Base2:
        def __init__(self, base2_param, **kwargs):
            super().__init__(**kwargs)
            self.base2_param = base2_param

    class Base3(mock_module):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--base3_option')

        def __init__(self, base3_param, **kwargs):
            super().__init__(**kwargs)
            self.base3_option = self.settings.base3_option
            self.base3_param = base3_param

    class Base4(mock_module):
        def __init__(self, base4_param, **kwargs):
            super().__init__(**kwargs)
            self.base4_param = base4_param

    class DerivedMulti(Base1, Base2, Base3, Base4):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--derived_option')

        def __init__(self, **kwargs):
            super().__init__(base1_param="ab",
                                               base2_param="cd",
                                               base3_param="ef",
                                               base4_param="gh",
                                               **kwargs)
            self.derived_option = self.settings.derived_option
            self.derived_member = 123

    local_settings = parse_settings(DerivedMulti, ["--base1_option=abc", "--base3_option=def", "--derived_option=ghi"])
    derived = construct_component(DerivedMulti, local_settings)
    assert derived.derived_option == "ghi"
    assert derived.base1_option == "abc"
    assert derived.base3_option == "def"
    assert derived.base1_param == "ab"
    assert derived.base2_param == "cd"
    assert derived.base3_param == "ef"
    assert derived.base4_param == "gh"


def test_construct_component_multiple_instance(mock_module):
    class Z(mock_module):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--zparam')

    class W(mock_module):
        dep = Dependency(Z)

        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--wparam')

        def __init__(self):
            super().__init__()
            if self.settings.wparam == "z":
                self.z = self.dep()
            else:
                self.nz = self.dep()

    settings1 = parse_settings(W, ["--wparam=z", "--zparam=abc"])
    w1 = construct_component(W, settings1)

    settings2 = parse_settings(W, ["--wparam=not_z", "--zparam=def"])
    w2 = construct_component(W, settings2)

    assert w1.settings.wparam == "z"
    assert w2.settings.wparam == "not_z"

    assert w1.z.settings.zparam == "abc"
    assert w2.nz.settings.zparam == "def"

    with pytest.raises(AttributeError) as exception_info:
        print(w1.nz.settings.zparam)
    assert "'W' object has no attribute 'nz'" in str(exception_info.value)

    with pytest.raises(AttributeError) as exception_info:
        print(w2.z.settings.zparam)
    assert "'W' object has no attribute 'z'" in str(exception_info.value)

    w1.settings.wparam = "new_value"
    assert w1.settings.wparam == "new_value"
    assert w2.settings.wparam == "not_z"


def test_settings_access_multiple_inheritance(mock_module):
    class BaseOne(mock_module):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--baseone')

        def set_baseone(self, value):
            self.settings.baseone = value

    class BaseTwo(mock_module):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--basetwo')

        def set_derived(self, value):
            self.settings.derived = value

    class Derived(BaseOne, BaseTwo):
        @staticmethod
        def define_arguments(parser):
            parser.add_argument('--derived')

    local_settings = parse_settings(Derived, ["--baseone=abc", "--basetwo=def", "--derived=ghi"])
    b1 = construct_component(BaseOne, local_settings)
    b2 = construct_component(BaseTwo, local_settings)
    d = construct_component(Derived, local_settings)
    assert d.settings.derived == "ghi"
    assert d.settings.baseone == "abc"
    assert d.settings.basetwo == "def"
    assert b1.settings.baseone == "abc"
    assert b2.settings.basetwo == "def"

    with pytest.raises(AttributeError) as exception_info:
        print(b1.settings.derived)
    assert "'BaseOne' object has no setting 'derived'" in str(exception_info.value)

    with pytest.raises(AttributeError) as exception_info:
        print(b2.settings.baseone)
    assert "'BaseTwo' object has no setting 'baseone'" in str(exception_info.value)

    b1.settings.baseone = "zyx"
    assert d.settings.baseone == "zyx"

    d.settings.basetwo = "lmn"
    assert b2.settings.basetwo == "lmn"

    d.set_baseone("value1")
    assert d.settings.baseone == "value1"

    d.set_derived("value2")
    assert d.settings.derived == "value2"

    with pytest.raises(AttributeError) as exception_info:
        b2.set_derived("value4")
    assert "'BaseTwo' object has no setting 'derived'" in str(exception_info.value)



import json
import urllib.parse

from .lib import utils
from .lib.ci_exception import CriticalCiException
from .lib.utils import make_block
from .modules.error_state import HasErrorState
from .modules.output import HasOutput
from .modules.structure_handler import HasStructure
from .modules.vcs.github_app_vcs import GithubToken


class GithubHandler(GithubToken, HasOutput, HasStructure, HasErrorState):

    @staticmethod
    def define_arguments(argument_parser):
        argument_parser.add_argument('--event', '-e', dest='event', metavar="GITHUB_EVENT",
                                     help='Currently parsed from "x-github-event" header of received web-hook')
        argument_parser.add_argument('--payload', '-pl', dest='payload', metavar="GITHUB_PAYLOAD",
                                     help='Full contents of web-hook received; pass a JSON file path, starting it '
                                          'with "@" character, or use an environment variable for convenience. '
                                          'Please note, that when passing a file, it is expected to be '
                                          'in UTF-8 encoding')
        argument_parser.add_argument('--trigger-url', '-tu', dest='trigger_url', metavar="TRIGGER_URL",
                                     help='URL for GET request to trigger the CI build and pass all parameters '
                                          'parsed from payload; if any constant parameters (like token) are '
                                          'required, include them in this string as well, e.g.: '
                                          '"http://jenkins.com/job/JobName/build?token=MYTOKEN"')
        argument_parser.add_argument('--suite-name', '-sn', dest='suite_name', metavar="SUITE_NAME", default="CI tests",
                                     help='GitHub check suite name, default is "CI tests"')
        argument_parser.add_argument('--target-repo', '-tr', dest='target_repo', metavar="TARGET_REPO",
                                     help='Repository name to process; leave blank to process any repo')
        argument_parser.add_argument('--verbose', '-v', dest='verbose', action="store_true",
                                     help='Show all params passed in URL (mostly for debug purposes)')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.check_required_option("event", """
            GitHub web-hook event is not specified.

            Please pass 'X-GitHub-Event' header contents of incoming web-hook request via
            command line parameter '--event' ('-e') or GITHUB_EVENT environment variable.
            """)
        self.check_required_option("trigger_url", """
            CI build trigger URL is not specified.

            Trigger URL is a string to be extended by parsed build parameters and used in a
            GET request to start a CI build that can report a GitHub build check. For
            example, 'http://jenkins.com/job/JobName/build?token=MYTOKEN'

            Please specify this parameter by using '--trigger-url' ('-tu') command line
            parameter or by setting TRIGGER_URL environment variable.
            """)
        self.payload = self.read_and_check_multiline_option("payload", """
            GitHub web-hook payload JSON is not specified.

            Please pass incoming web-hook request payload to this parameter directly via
            '--payload' ('-pl') command line parameter or by setting GITHUB_PAYLOAD
            environment variable, or by passing file path as the argument value (start
            filename with '@' character, e.g. '@/tmp/file.json' or '@payload.json' for
            relative path starting at current directory). Please note, that when passing
            a file, it's expected to be in UTF-8 encoding
            """)

    @make_block("Analysing trigger payload")
    def execute(self):
        try:
            payload = json.loads(self.payload)
        except json.decoder.JSONDecodeError as error:
            text = f"Provided payload value could not been parsed as JSON and returned the following error:\n {error}"
            raise CriticalCiException(text) from error

        try:
            if self.settings.target_repo:
                parsed_repo = payload['repository']['full_name']
                if parsed_repo != self.settings.target_repo:
                    text = f"This payload has come from '{parsed_repo}' repository, while due to parameter passed to "
                    text += f" '--target-repo' only '{self.settings.target_repo}' repository is handled. Skipping..."
                    self.out.log(text)
                    return

            if self.settings.event == "check_suite" and (payload["action"] in ["requested", "rerequested"]):
                url = payload["repository"]["url"] + "/check-runs"
                data = {"name": self.settings.suite_name, "head_sha": payload["check_suite"]["head_sha"]}
                headers = {'Authorization': f"token {self.get_token(payload['installation']['id'])}",
                           'Accept': 'application/vnd.github.antiope-preview+json'}
                self.out.log(f"Sending request to {url}")
                if self.settings.verbose:
                    self.out.log(f"Headers are:\n{headers}\nOther passed params are:\n{data}")
                utils.make_request(url=url, request_method="POST", json=data, headers=headers, timeout=5*60)
                self.out.log("Successfully created check run")

            elif self.settings.event == "check_run" and \
                    (payload["action"] in ["requested", "rerequested", "created"]) and \
                    (str(payload["check_run"]["app"]["id"]) == str(self.settings.integration_id)):
                param_dict = {
                    "GIT_REFSPEC": payload["check_run"]["check_suite"]["head_branch"],
                    "GIT_CHECKOUT_ID": payload["check_run"]["head_sha"],
                    "GITHUB_CHECK_ID": payload["check_run"]["id"],
                    "GIT_REPO": payload["repository"]["clone_url"],
                    "GITHUB_INSTALLATION_ID": payload["installation"]["id"]
                }
                self.out.log(f"Triggering {urllib.parse.urljoin(self.settings.trigger_url, '?...')}")
                response = utils.make_request(self.settings.trigger_url, params=param_dict)
                if self.settings.verbose:
                    self.out.log(f"Triggered {response.url}")
                else:
                    self.out.log("Successfully triggered")

            else:
                self.out.log("Unhandled event, skipping...")
        except KeyError as error:
            raise CriticalCiException(f"Could not find key {error} in provided payload:\n{payload}") from error
        except TypeError as error:
            raise CriticalCiException("Parsed payload JSON does not correspond to expected format") from error

    def finalize(self):
        pass



# pylint: disable = redefined-outer-name

from typing import Optional
import pathlib
import pytest

from .conftest import FuzzyCallChecker
from .git_utils import GitTestEnvironment, GitClient, GitServer
from .utils import HttpChecker


@pytest.fixture()
def git_poll_environment(git_client: GitClient, tmp_path: pathlib.Path):
    yield GitTestEnvironment(git_client, tmp_path, test_type="poll")


def make_branch_with_changes(git_server: GitServer, branch_name: str, commits_number: int,
                             branch_from: Optional[str] = None):
    """
    Creates a branch from the current or specified (by name) and adds passed commits number.

    :param git_server: Reference to the instance of the GitServer class from the confest.py
    :param branch_name: Name of newly created branch
    :param commits_number: Number of commits to be made in created branch
    :param branch_from: Branch name to check out from. By default branch will be made from the current
    :return: A list of commits "hashes"
    """
    if branch_from:
        git_server.switch_branch(branch_from)
    git_server.make_branch(branch_name)
    commits = [git_server.commit_new_file() for _ in range(commits_number)]
    git_server.switch_branch(git_server.target_branch)
    return commits


def assert_polled_commits(commits: list, stdout_checker: FuzzyCallChecker, collected_http: HttpChecker) -> None:
    for commit in commits:
        stdout_checker.assert_has_calls_with_param("==> Detected commit " + commit)
        collected_http.assert_request_was_made({"cl": [commit]})


def test_max_number_commits(stdout_checker: FuzzyCallChecker, git_poll_environment: GitTestEnvironment):
    # initialize working directory with initial data
    git_poll_environment.run_with_http_server()

    # ACT
    # make changes in polled branch
    allowed_commits_number: int = git_poll_environment.settings.Poll.max_number
    changes_to_polled = [git_poll_environment.server.commit_new_file() for _ in range(allowed_commits_number + 1)]

    # ASSERT
    # run poll again and trigger the url twice
    collected_http: HttpChecker = git_poll_environment.run_with_http_server()
    assert_polled_commits(changes_to_polled[1:],
                          stdout_checker,
                          collected_http)
    # Ensure that oldest commit is beyond "allowed_commits_number" and is not polled
    stdout_checker.assert_absent_calls_with_param("==> Detected commit " + changes_to_polled[0])


def test_merge_one_branch_ff(stdout_checker: FuzzyCallChecker, git_poll_environment: GitTestEnvironment):
    # initialize working directory with initial data
    git_poll_environment.run_with_http_server()

    # ACT
    # make changes in polled branch
    change_to_polled = git_poll_environment.server.commit_new_file()
    # make a branch from polled and add 2 new commits
    changes_to_branch = make_branch_with_changes(git_poll_environment.server, "test_branch", 2)
    # merge created branch to the polled using fast-forward option
    git_poll_environment.server.merge_branch("test_branch", fast_forward=True)

    # ASSERT
    # run poll again and trigger the url twice
    collected_http: HttpChecker = git_poll_environment.run_with_http_server()
    assert_polled_commits([change_to_polled] + changes_to_branch,
                          stdout_checker,
                          collected_http)


def test_merge_one_branch_noff(stdout_checker: FuzzyCallChecker, git_poll_environment: GitTestEnvironment):
    # initialize working directory with initial data
    git_poll_environment.run_with_http_server()

    # ACT
    # make changes in polled branch
    change_to_polled = git_poll_environment.server.make_a_change()
    # make a branch from polled and add 2 new commits
    changes_to_branch = make_branch_with_changes(git_poll_environment.server, "test_branch", 2)
    # merge created branch to the polled using fast-forward option
    git_poll_environment.server.merge_branch("test_branch", fast_forward=False)
    merge_commit_id = git_poll_environment.server.get_last_commit()

    # ASSERT
    # run poll again and trigger the url twice
    collected_http: HttpChecker = git_poll_environment.run_with_http_server()
    assert_polled_commits([change_to_polled] + [merge_commit_id],
                          stdout_checker,
                          collected_http)
    stdout_checker.assert_absent_calls_with_param("==> Detected commit " + changes_to_branch[0])
    stdout_checker.assert_absent_calls_with_param("==> Detected commit " + changes_to_branch[1])


def test_merge_two_subsequent_branches_noff(stdout_checker: FuzzyCallChecker, git_poll_environment: GitTestEnvironment):
    # initialize working directory with initial data
    git_poll_environment.run_with_http_server()

    # ACT
    # make changes in polled branch
    change_to_polled = git_poll_environment.server.make_a_change()
    # make a branch from polled and add 1 new commit
    changes_to_first_branch = make_branch_with_changes(git_poll_environment.server, "test_branch1", 1)
    # make a branch from test_branch1 and add 1 new commit
    changes_to_second_branch = make_branch_with_changes(git_poll_environment.server, "test_branch2", 1)
    # merge created branch to the polled using fast-forward option
    git_poll_environment.server.merge_branch("test_branch2", fast_forward=False)
    merge_commit_id = git_poll_environment.server.get_last_commit()

    # ASSERT
    # run poll again and trigger the url twice
    collected_http: HttpChecker = git_poll_environment.run_with_http_server()
    assert_polled_commits([change_to_polled] + [merge_commit_id],
                          stdout_checker,
                          collected_http)
    stdout_checker.assert_absent_calls_with_param("==> Detected commit " + changes_to_first_branch[0])
    stdout_checker.assert_absent_calls_with_param("==> Detected commit " + changes_to_second_branch[0])


def test_merge_two_subsequent_branches_ff(stdout_checker: FuzzyCallChecker, git_poll_environment: GitTestEnvironment):
    # initialize working directory with initial data
    git_poll_environment.run_with_http_server()

    # ACT
    # make changes in polled branch
    change_to_polled = git_poll_environment.server.commit_new_file()
    # make a branch from polled and add 1 new commit
    changes_to_first_branch = make_branch_with_changes(git_poll_environment.server, "test_branch1", 1)
    # make a branch from test_branch and add 1 new commit
    changes_to_second_branch = make_branch_with_changes(git_poll_environment.server, "test_branch2", 1,
                                                        branch_from="test_branch1")
    # merge created branch to the polled using fast-forward option
    git_poll_environment.server.merge_branch("test_branch2", fast_forward=True)

    # ASSERT
    # run poll again and trigger the url twice
    collected_http: HttpChecker = git_poll_environment.run_with_http_server()
    assert_polled_commits([change_to_polled] + changes_to_first_branch + changes_to_second_branch,
                          stdout_checker,
                          collected_http)


def test_merge_one_branch_noff_1_commit_behind(stdout_checker: FuzzyCallChecker, git_poll_environment: GitTestEnvironment):
    # initialize working directory with initial data
    git_poll_environment.run_with_http_server()

    # ACT
    # make a branch from polled
    git_poll_environment.server.make_branch("test_branch")
    # make change to the polled branch
    git_poll_environment.server.switch_branch(git_poll_environment.server.target_branch)
    change_to_polled = git_poll_environment.server.commit_new_file()
    # make change to created test branch
    git_poll_environment.server.switch_branch("test_branch")
    changes_to_branch = git_poll_environment.server.commit_new_file()
    # merge test branch to the polled
    git_poll_environment.server.switch_branch(git_poll_environment.server.target_branch)
    git_poll_environment.server.merge_branch("test_branch", fast_forward=False)
    merge_commit_id = git_poll_environment.server.get_last_commit()

    # ASSERT
    # run poll again and trigger the url twice
    collected_http: HttpChecker = git_poll_environment.run_with_http_server()
    assert_polled_commits([change_to_polled, merge_commit_id],
                          stdout_checker,
                          collected_http)
    stdout_checker.assert_absent_calls_with_param("==> Detected commit " + changes_to_branch)


@pytest.mark.xfail
def test_merge_ff_commit_merged_from_polled(stdout_checker: FuzzyCallChecker, git_poll_environment: GitTestEnvironment):
    # initialize working directory with initial data
    git_poll_environment.run_with_http_server()

    # ACT
    # make a branch from polled
    git_poll_environment.server.make_branch("test_branch")
    # make a change to the polled branch
    git_poll_environment.server.switch_branch(git_poll_environment.server.target_branch)
    change_to_polled = git_poll_environment.server.make_a_change()
    # make a change to created test branch
    git_poll_environment.server.switch_branch("test_branch")
    change_to_branch = git_poll_environment.server.commit_new_file()
    # merge polled branch to the current
    # fast-forward is not possible in this case
    git_poll_environment.server.merge_branch(git_poll_environment.server.target_branch, fast_forward=False)
    merge_commit_to_branch = git_poll_environment.server.get_last_commit()
    # merge test branch to the polled using fast forward
    git_poll_environment.server.switch_branch(git_poll_environment.server.target_branch)
    git_poll_environment.server.merge_branch("test_branch", fast_forward=True)

    # ASSERT
    # run poll again and trigger the url twice
    collected_http: HttpChecker = git_poll_environment.run_with_http_server()
    assert_polled_commits([change_to_polled, change_to_branch],
                          stdout_checker,
                          collected_http)
    stdout_checker.assert_absent_calls_with_param("==> Detected commit " + merge_commit_to_branch)



import locale
import sys
from typing import List

from .base_output import BaseOutput

__all__ = [
    "TerminalBasedOutput"
]


class Colors:
    red = "\033[1;31m"
    dark_red = "\033[0;31m"
    green = "\033[1;32m"
    blue = "\033[1;34m"
    dark_cyan = "\033[0;36m"
    dark_yellow = "\033[0;33m"
    reset = "\033[00m"


class TerminalBasedOutput(BaseOutput):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.block_level = 0
        self.unicode_acceptable = (locale.getpreferredencoding() == "UTF-8")  # pylint: disable = superfluous-parens

    @staticmethod
    def _stdout(*args, **kwargs) -> None:
        sys.stdout.write(''.join(args))
        if not kwargs.get("no_enter", False):
            sys.stdout.write('\n')

    def _indent(self) -> None:
        for x in range(0, self.block_level):
            self._stdout("  " * x, " |   ", no_enter=True)

    def _print_lines(self, *args, **kwargs) -> None:
        result: str = ''.join(args)
        lines: List[str] = result.splitlines(False)
        for line in lines:
            self._indent()
            self._stdout(line)

    def log(self, line: str) -> None:
        self._print_lines("==> ", line)

    def log_error(self, description: str) -> None:
        self._print_lines(Colors.dark_red, "Error: ", Colors.reset, description)

    def log_external_command(self, command: str) -> None:
        self._print_lines("$ ", command)

    def log_stdout(self, line: str) -> None:
        self._print_lines(line)

    def log_stderr(self, line: str) -> None:
        self._print_lines(Colors.dark_yellow, "stderr: ", Colors.reset, line)

    def open_block(self, num_str: str, name: str) -> None:
        self._indent()
        self._stdout(num_str, ' ', Colors.blue, name, Colors.reset)
        self.block_level += 1

    def close_block(self, num_str: str, name: str, status: str) -> None:
        self.block_level -= 1
        self._indent()
        if self.unicode_acceptable:
            block_end = " \u2514 "
        else:
            block_end = " | "

        if status == "Failed":
            self._stdout(self.block_level * "  ", block_end, Colors.red, "[Failed]", Colors.reset)
        else:
            self._stdout(self.block_level * "  ", block_end, Colors.green, "[Success]", Colors.reset)
        self._indent()
        self._stdout()

    def log_skipped(self, message: str) -> None:
        self._print_lines(Colors.dark_cyan, message, Colors.reset)

    def log_summary_step(self, step_title: str, has_children: bool, status: str) -> None:
        color = Colors.red
        if status.lower() == "success":
            color = Colors.green

        if not has_children:
            step_title += " - " + color + status + Colors.reset

        self.log(step_title)

    def report_build_problem(self, description: str) -> None:
        pass

    def set_build_status(self, status: str) -> None:
        pass



from ...lib.gravity import Module


class BaseOutput(Module):
    """
    Abstract base class for output drivers.  Defines methods that have to be implemented by each module that
    implements a type of output.
    """

    def log_execution_start(self, title: str, version: str) -> None:
        """
        Print the universum startup message, for example:
        "==> Universum xx.yy.zz started execution"

        :param title: name of the product
        :param version: version of the product
        """
        self.log(self._build_execution_start_msg(title, version))

    def log_execution_finish(self, title: str, version: str) -> None:
        """
        Print the universum finished message, for example:
        "==> Universum xx.yy.zz finished execution"

        :param title: name of the product
        :param version: version of the product
        """
        self.log(self._build_execution_finish_msg(title, version))

    def log(self, line: str) -> None:
        """
        Print a line to the output. The most generic type of log.
        Mostly used for printing own messages of universum.

        :param line: the line to print
        """
        raise NotImplementedError

    def log_error(self, description: str) -> None:
        """
        Print an error message to the output. Error may use different colors or other formatting.
        Used for printing own error messages of the universum as well as error messages from other modules and libraries
        and operating system.

        :param description: the detailed error message
        """
        raise NotImplementedError

    def log_external_command(self, command: str) -> None:
        """
        Print the start of the launch of the external command. Full command line with all arameters is supposed to be
        printed.

        :param command: external command with parameters
        """
        raise NotImplementedError

    def log_stdout(self, line: str) -> None:
        """
        Print one line of the standard output of the previously launched external command. The only usage is for
        printing output of the commands that are launched during the execution of the universum config.

        :param line: the line to print
        """
        raise NotImplementedError

    def log_stderr(self, line: str) -> None:
        """
        Print one line of the standard error output of the previously launched external command. The only usage is for
        printing standard error output of the commands that are launched during the execution of the universum config.

        :param line: the line to print
        """
        raise NotImplementedError

    def open_block(self, num_str: str, name: str) -> None:
        """
        Start the block of the output. Block is a group of related output lines.
        Depending on the capabilities of the output, the blocks can be marked by indentation, color or numbering.
        Output can also implement dynamic blocks collapsing and expanding.

        :param num_str: the string that represents the block number within the hierarchy.
            Inner blocks are numbered using dot notation.
            The number string is passed as a separate parameter to allow implementing different formatting for number and
            name.
        :param name: the title of the block.
        """
        raise NotImplementedError

    def close_block(self, num_str: str, name: str, status: str) -> None:
        """
        End the block of the output.

        :param num_str: the string that represents the block number
        :param name: the title of the block
        :param status: the completion status of the block: "Passed" or "Failed"
        """
        raise NotImplementedError

    def log_skipped(self, message: str) -> None:
        """
        Print a message related to skipping steps. It could be a warning that further steps are skipped, or a message
        that the step is skipped.

        :param message: the message to print
        """
        raise NotImplementedError

    def log_summary_step(self, step_title: str, has_children: bool, status: str) -> None:
        """
        Print a summary of the step execution at the end of the run.

        :param step_title: the title of the step
        :param has_children: whether the step has children or not. Typically, if step has children, the status is not
            included in the summary.
        :param status: the status of the step: "Passed", "Failed" or "Skipped"
        """
        raise NotImplementedError

    def report_build_problem(self, description: str) -> None:
        """
        TeamCity only. Report run problem to the TeamCity via the output. Also fails the build regardless of the process
        exit code. Multiple build problems for single run can be reported.

        :param description: the description of the problem
        """
        raise NotImplementedError

    def set_build_status(self, status: str) -> None:
        """
        TeamCity only. Set the title of the build. Changes the status of the build visible in the configuration build
        history and on the build page itself. For builds containing build problems, by default the status is
        automatically populated with them by the TeamCity server itself. This method allows to override the status.

        :param status: the status (title) of the build
        """
        raise NotImplementedError

    @staticmethod
    def _build_execution_start_msg(title: str, version: str) -> str:
        return f"{title} {version} started execution"

    @staticmethod
    def _build_execution_finish_msg(title: str, version: str) -> str:
        return f"{title} {version} finished execution"



# pylint: disable = redefined-outer-name

import os
import re
from datetime import datetime
from enum import Enum, auto
import colorsys
import pytest

from selenium import webdriver
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.firefox.webelement import FirefoxWebElement

from universum import __main__
from . import utils


config = r"""
from universum.configuration_support import Configuration

failed_step_cmd = ["bash", "-c", '1>&2 echo "error"; exit 1']

success_links_step = Configuration([dict(name="Success links step",
    command=["echo", "http://www.samsung.com/ https://www.samsung.com/ ftp://www.samsung.com " +
        r"file://www.samsung.com/remote_file file:///local\ file mailto:asdf@samsung.com"])])
success_ansi_step = Configuration([dict(name="Success ansi step",
    command=["echo", "\u001b[31m red \u001b[0m, \u001b[42m\u001b[34m blue on green \u001b[0m, " +
        "\u001b[1m\u001b[4m\u001b[36m bold underline cyan \u001b[0m"])])
failed_step = Configuration([dict(name="Failed step", command=failed_step_cmd)])
partially_success_step = Configuration([dict(name="Partially success step: ")])
all_success_step = Configuration([dict(name="All success step: ")])
all_failed_step = Configuration([dict(name="All failed step: ")])
failed_critical_step = Configuration([dict(name="Failed step", command=failed_step_cmd, critical=True)])

configs = \
    success_links_step + \
    success_ansi_step + \
    failed_step + \
    partially_success_step * (success_links_step + failed_step) + \
    all_success_step * (success_links_step + success_ansi_step) + \
    all_failed_step * (failed_step + failed_step) + \
    failed_critical_step + success_links_step
"""

log_name = "universum_log"


def create_environment(test_type, tmp_path):
    env = utils.LocalTestEnvironment(tmp_path, test_type)
    env.configs_file.write_text(config)
    env.settings.Output.html_log = log_name
    return env


@pytest.fixture(params=["main", "nonci"])
def environment_main_and_nonci(request, tmp_path):
    yield create_environment(request.param, tmp_path)


@pytest.fixture()
def browser():
    options = Options()
    options.headless = True
    firefox = webdriver.Firefox(options=options)
    yield firefox
    firefox.close()


def test_cli_log_custom_name(tmp_path):
    custom_log_name = "custom_name.html"
    artifact_dir = check_cli(tmp_path, ["-hl", custom_log_name])
    assert os.path.exists(os.path.join(artifact_dir, custom_log_name))


def test_cli_log_default_name(tmp_path):
    artifact_dir = check_cli(tmp_path, ["-hl"])
    assert os.path.exists(os.path.join(artifact_dir, "universum_log.html"))


def test_cli_no_log_requested(tmp_path):
    artifact_dir = check_cli(tmp_path, [])
    for file_name in os.listdir(artifact_dir):
        assert not file_name.endswith(".html")


def test_success(environment_main_and_nonci, browser):
    environment_main_and_nonci.run()
    check_html_log(environment_main_and_nonci.artifact_dir, browser)


def test_success_clean_build(tmp_path, browser):
    env = create_environment("main", tmp_path)
    env.settings.Main.clean_build = True
    env.run()
    check_html_log(env.artifact_dir, browser)


def check_html_log(artifact_dir, browser):
    log_path = os.path.join(artifact_dir, log_name)
    assert os.path.exists(log_path)

    browser.get(f"file://{log_path}")
    html_body = TestElement.create(browser.find_element_by_tag_name("body"))
    body_elements = html_body.find_elements_by_xpath("./pre")
    assert len(body_elements) == 1

    universum_log_element = TestElement.create(body_elements[0])
    steps_section = universum_log_element.get_section_by_name("Executing build steps")
    steps_body = steps_section.get_section_body()

    check_initial_sections_display_state(universum_log_element, steps_body)
    check_manual_section_expanding(steps_body)
    check_sections_indentation(steps_section)
    check_coloring(html_body, universum_log_element, steps_body)
    check_links_wrapping(steps_body)
    check_ansi_colors_convertion(steps_body)
    check_dark_mode(html_body, universum_log_element)
    check_timestamps(html_body, universum_log_element)


def check_initial_sections_display_state(universum_log_element, steps_body):
    assert steps_body.is_displayed()
    sections_expanded_state = {
        "Success links step": False,
        "Failed step": True,
        "Partially success step": True,
        "Partially success step: Success links step": False,
        "Partially success step: Failed step": True,
        "All success step": False,
        "All failed step": True,
        "All failed step: Failed step": True
    }
    for section_name, expanded_state in sections_expanded_state.items():
        assert steps_body.get_section_body_by_name(section_name).is_displayed() == expanded_state
    assert universum_log_element.get_section_body_by_name("Reporting build result").is_displayed()


def check_manual_section_expanding(steps_body):
    success_step = steps_body.get_section_by_name("Success links step")
    success_step_body = success_step.get_section_body()
    assert not success_step_body.is_displayed()
    success_step.click()
    assert success_step_body.is_displayed()
    success_step.click()
    assert not success_step_body.is_displayed()


def check_sections_indentation(steps_section):
    steps_body = steps_section.get_section_body()
    step_lvl1_first = steps_body.get_section_by_name("Failed step")
    step_lvl1_second = steps_body.get_section_by_name("Partially success step")
    assert step_lvl1_first.indent == step_lvl1_second.indent

    step_lvl1_body = step_lvl1_second.get_section_body()
    step_lvl2_first = step_lvl1_body.get_section_by_name("Success links step")
    step_lvl2_second = step_lvl1_body.get_section_by_name("Failed step")
    assert step_lvl2_first.indent == step_lvl2_second.indent

    assert steps_section.indent < step_lvl1_first.indent < step_lvl2_first.indent


def check_coloring(body_element, universum_log_element, steps_body):
    check_body_coloring(body_element)
    check_title_and_status_coloring(steps_body)
    check_skipped_steps_coloring(steps_body)
    check_steps_report_coloring(universum_log_element)
    check_errors_tags_coloring(steps_body)


def check_body_coloring(body_element):
    assert body_element.color == Color.BLACK
    assert body_element.background_color == Color.WHITE


def check_title_and_status_coloring(steps_body):
    check_section_coloring(steps_body.get_section_by_name("Success links step"))
    check_section_coloring(steps_body.get_section_by_name("Failed step"), is_failed=True)
    check_section_coloring(steps_body.get_section_by_name("Partially success step"), is_failed=True)

    composite_step_body = steps_body.get_section_body_by_name("Partially success step")
    check_section_coloring(composite_step_body.get_section_by_name("Success links step"))
    check_section_coloring(composite_step_body.get_section_by_name("Failed step"), is_failed=True)


def check_skipped_steps_coloring(steps_body):
    skipped_steps = steps_body.find_elements_by_class_name("skipped")
    assert skipped_steps
    skipped_steps = [TestElement.create(step) for step in skipped_steps]
    for step in skipped_steps:
        assert step.color == Color.CYAN


def check_steps_report_coloring(universum_log_element):
    report_section_body = universum_log_element.get_section_body_by_name("Reporting build result")
    xpath_selector = "./*[text() = 'Success' or text() = 'Failed' or text() = 'Skipped']"
    elements = [TestElement.create(el) for el in report_section_body.find_elements_by_xpath(xpath_selector)]
    assert elements
    for el in elements:
        if el.text == "Success":
            assert el.color == Color.GREEN
        elif el.text in ("Failed", "Skipped"):
            assert el.color == Color.RED
        else:
            assert False, f"Unexpected element text: '{el.text}'"


def check_errors_tags_coloring(steps_body):
    step_body = steps_body.get_section_body_by_name("Failed step")
    exception_tag = step_body.get_child_by_text("Error:")
    assert exception_tag.color == Color.RED
    stderr_tag = step_body.get_child_by_text("stderr:")
    assert stderr_tag.color == Color.YELLOW


def check_section_coloring(step, is_failed=False):
    check_text_item_style(step.get_section_title(), is_failed, normal_color=Color.BLUE)
    if not is_failed:
        step.click()  # open section body
    check_text_item_style(step.get_section_status(), is_failed, normal_color=Color.GREEN)
    if not is_failed:
        step.click()  # close section body


def check_text_item_style(item, is_failed, normal_color):
    exp_color = Color.RED if is_failed else normal_color
    assert item.color == exp_color
    check_text_is_bold(item)


def check_text_is_bold(element):
    font_weight_bold = "700"
    assert element.font_weight == font_weight_bold


def check_links_wrapping(steps_body):
    step = steps_body.get_section_by_name("Success links step")
    body = step.get_section_body()
    assert not body.is_displayed()
    step.click()
    for url_scheme in ("http", "https", "ftp", "mailto"):
        assert body.find_element_by_partial_link_text(url_scheme)
    file_links = body.find_elements_by_partial_link_text("file")
    assert len(file_links) == 4 # `echo` parameters also counts
    link_with_whitespace = file_links[1]
    assert r"\ " in link_with_whitespace.text
    assert "%20" in link_with_whitespace.get_attribute("href")
    step.click() # restore section closed state


def check_ansi_colors_convertion(steps_body):
    step = steps_body.get_section_by_name("Success ansi step")
    body = step.get_section_body()
    assert not body.is_displayed()
    step.click()

    span_elements = body.find_elements_by_tag_name("span")
    def is_target_span_element(element):
        return element.text and element.get_attribute("style")
    span_elements = [TestElement.create(el) for el in span_elements if is_target_span_element(el)]
    assert len(span_elements) == 6 # 3 printed elements + 3 `echo` argument elements

    for element in span_elements:
        if "red" in element.text:
            assert element.color == Color.RED
        elif "blue on green" in element.text:
            assert element.color == Color.BLUE
            assert element.background_color == Color.GREEN
        elif "bold underline cyan" in element.text:
            check_text_is_bold(element)
            assert "underline" in element.value_of_css_property("text-decoration")
            assert element.color == Color.CYAN
        else:
            assert False, f"Unexpected span text: {element.text}"

    step.click() # restore section closed state


def check_dark_mode(body_element, universum_log_element):
    dark_mode_switch = TestElement.create(body_element.find_element_by_xpath("./label[@for='dark-checkbox']"))
    dark_mode_switch.click()
    assert universum_log_element.color == Color.WHITE
    assert universum_log_element.background_color == Color.BLACK
    dark_mode_switch.click()
    assert universum_log_element.color == Color.BLACK


def check_timestamps(body_element, universum_log_element):
    timestamp_element = universum_log_element.find_element_by_xpath("./*[@class='time']")
    assert not timestamp_element.is_displayed()

    time_switch = TestElement.create(body_element.find_element_by_xpath("./label[@for='time-checkbox']"))
    time_switch.click()
    assert timestamp_element.is_displayed()

    timestamp = datetime.strptime(timestamp_element.text, "%Y-%m-%d %H:%M:%S")
    delta = abs(datetime.now() - timestamp)
    assert delta.days == 0
    assert delta.seconds <= 60


def check_cli(tmp_path, html_log_params):
    artifact_dir = tmp_path / "artifacts"
    config_file = tmp_path / "configs.py"
    config_file.write_text(config, "utf-8")

    cli_params = ["-vt", "none",
                  "-fsd", str(tmp_path),
                  "-cfg", str(config_file),
                  "-ad", str(artifact_dir)]
    html_log_params.extend(cli_params)
    result = __main__.main(html_log_params)

    assert result == 0
    return artifact_dir


class TestElement(FirefoxWebElement):

    __test__ = False # disable pytest collection for this class

    @staticmethod
    def create(element):
        assert element
        element.__class__ = TestElement
        return element

    def get_section_body_by_name(self, section_name):
        return self.get_section_by_name(section_name).get_section_body()

    # <any_tag>  <-- self
    #     <input type="checkbox" id="1." class="hide">
    #     <label for="1.">  <-- returning this element
    #         <span class="sectionLbl">1. Section name</span>  <-- Searching for this element
    #     </label>
    #     <div>Section body</div>
    # </any_tag>
    def get_section_by_name(self, section_name):
        span_elements = self.find_elements_by_class_name("sectionLbl")
        result = None
        for el in span_elements:
            if section_name in el.text:
                return TestElement.create(el.find_element_by_xpath(".."))
        return TestElement.create(result)

    # <input type="checkbox" id="1." class="hide">
    # <label for="1.">  <-- self
    #     <span class="sectionLbl">1. Section name</span>
    # </label>
    # <div>Section body</div>  <-- returning this element
    def get_section_body(self):
        body = TestElement.create(self.find_element_by_xpath("./following-sibling::*[1]"))
        assert body.tag_name == "div"
        return body

    # <label for="1.">  <-- self
    #     <span class="sectionLbl">
    #         <span class="sectionTitle">1. Section name</span>  <-- returning this element
    #     </span>
    # </label>
    def get_section_title(self):
        label_span = self.find_element_by_class_name("sectionLbl")
        assert label_span
        title_span = label_span.find_element_by_tag_name("span")
        return TestElement.create(title_span)

    # <label for="1.">  <-- self
    #     ...
    # </label>
    # <div>
    #     ...
    #     <span class="sectionSuccessStatus">[Success]</span>  <-- returning this element
    # </div>
    def get_section_status(self):
        body = self.get_section_body()
        xpath_selector = "./*[contains(text(), '[Success]') or contains(text(), '[Failed]')]"
        return TestElement.create(body.find_elements_by_xpath(xpath_selector)[-1])

    def get_child_by_text(self, text):
        xpath_selector = f"./*[text() = '{text}']"
        return TestElement.create(self.find_element_by_xpath(xpath_selector))

    @property
    def indent(self):
        return self.location['x']

    @property
    def color(self):
        return self._get_primary_color(self.value_of_css_property("color"))

    @property
    def background_color(self):
        return self._get_primary_color(self.value_of_css_property("background-color"))

    @property
    def font_weight(self):
        return self.value_of_css_property("font-weight")

    @staticmethod
    def _get_primary_color(rgb_str):
        re_result = re.match(r"^rgb\((\d{1,3}), (\d{1,3}), (\d{1,3})\)$", rgb_str)
        assert re_result
        red, green, blue = int(re_result.group(1)), int(re_result.group(2)), int(re_result.group(3))
        hue, lightness, _ = colorsys.rgb_to_hls(red/255.0, green/255.0, blue/255.0)
        hue = 360 * hue
        lightness = 100 * lightness

        color = None
        if lightness <= 20:
            color = Color.BLACK
        elif lightness >= 80:
            color = Color.WHITE
        elif 0 <= hue <= 30 or 330 <= hue <= 360:
            color = Color.RED
        elif 30 <= hue <= 80:
            color = Color.YELLOW
        elif 80 <= hue <= 150:
            color = Color.GREEN
        elif 150 <= hue <= 190:
            color = Color.CYAN
        elif 190 <= hue <= 270:
            color = Color.BLUE
        elif 270 <= hue <= 330:
            color = Color.PURPLE
        else:
            assert False, "Should not occur"
        return color


class Color(Enum):
    BLACK = auto()
    WHITE = auto()
    RED = auto()
    GREEN = auto()
    BLUE = auto()
    YELLOW = auto()
    PURPLE = auto()
    CYAN = auto()



from .deployment_utils import UniversumRunner

config = """
from universum.configuration_support import Configuration

configs = Configuration([dict(name="artifact check",
                              command=["bash", "-c", '''cat {}/test_nonci.txt''']),
#                                                    ^ this helps to check artifact is deleted before launch 

                         dict(name="test_step", artifacts="test_nonci.txt",
                              command=["bash", "-c", '''echo "pwd:[$(pwd)]" && echo "test nonci" > test_nonci.txt'''])])
#                                                    ^ this helps to check the path is not changed
"""


def test_launcher_output(docker_nonci: UniversumRunner):
    """
    This test verifies that nonci mode changes the behavior of the universum by checking its output to console and log
    files. Specifically, it checks the following features of the nonci mode:
     - default output of the step is console, even when running in the terminal
     - sources are not copied to temp directory
     - artifacts are deleted before launching configs
     - version control and review system are not used
     - project root is set to current directory
    """
    cwd = str(docker_nonci.local.root_directory.absolute())
    artifacts = docker_nonci.artifact_dir
    file_output_expected = f"Adding file {artifacts}/test_step_log.txt to artifacts"
    pwd_string_in_logs = f"pwd:[{cwd}]"

    docker_nonci.environment.assert_successful_execution(
        f"bash -c 'mkdir \"{artifacts}\"; echo \"Old artifact\" > \"{artifacts}/test_nonci.txt\"'")

    docker_nonci.project_root = None  # type: ignore
    console_out_log = docker_nonci.run(config.format(artifacts), workdir=cwd)

    # the following logs are only present in the default mode of the universum
    assert file_output_expected not in console_out_log          # nonci doesn't write logs to the file by default
    assert "Reporting build start" not in console_out_log       # nonci doesn't report build start
    assert "Copying sources" not in console_out_log             # nonci doesn't copy sources
    assert "Cleaning copied sources" not in console_out_log     # nonci doesn't delete sources after work is done

    # the following logs are specific to the nonci mode of the universum
    assert "Cleaning artifacts..." in console_out_log           # nonci cleans artifacts on project launch
    assert "Old artifact" not in console_out_log                # the artifacts are actually deleted
    assert pwd_string_in_logs in console_out_log                # nonci launches step in the same directory

    # nonci doesn't require to clean artifacts between calls
    log = docker_nonci.run(config.format(artifacts), additional_parameters='-lo file', workdir=cwd)
    assert file_output_expected in log

    assert console_out_log != log
    step_log = docker_nonci.environment.assert_successful_execution(
        f"cat '{artifacts}/test_step_log.txt'")
    assert pwd_string_in_logs in step_log

    # second call of universum must not contain previous step log
    docker_nonci.run("""
from universum.configuration_support import Configuration

configs = Configuration([dict(name="test_step",
                              command=["bash", "-c", '''echo "Separate run"'''])])
""", additional_parameters='-lo file', workdir=cwd)

    second_run_step_log = docker_nonci.environment.assert_successful_execution(
        f"cat '{artifacts}/test_step_log.txt'")
    assert pwd_string_in_logs not in second_run_step_log
    assert "Separate run" in second_run_step_log


def test_custom_artifact_dir(docker_nonci: UniversumRunner):
    docker_nonci.run(config, additional_parameters='-ad ' + '/my/artifacts/')
    docker_nonci.environment.assert_successful_execution("test -f /my/artifacts/test_nonci.txt")



import json
import urllib.parse
from typing import Dict, List, Union, Optional
from typing_extensions import Self

from . import git_vcs
from ..reporter import ReportObserver, Reporter
from ...lib import utils
from ...lib.gravity import Dependency
from ...lib.module_arguments import ModuleArgumentParser

__all__ = [
    "GithubActionsMainVcs"
]


class GithubActionsMainVcs(ReportObserver, git_vcs.GitMainVcs):
    """
    This class mostly contains functions for GitHub report observer
    """
    reporter_factory: Dependency = Dependency(Reporter)

    @staticmethod
    def define_arguments(argument_parser: ModuleArgumentParser) -> None:
        parser: ModuleArgumentParser = argument_parser.get_or_create_group("GitHub Actions",
                                                                           "GitHub repository settings for GH Actions")

        parser.add_argument("--ghactions-token", "-ght", dest="token", metavar="GITHUB_TOKEN",
                            help="Is stored in ${{ secrets.GITHUB_TOKEN }}")
        parser.add_argument("--ghactions-payload", "-ghp", dest="payload", metavar="GITHUB_PAYLOAD",
                            help="File path: ${{ github.event_path }}")

    def __init__(self, *args, **kwargs) -> None:
        self.settings.repo = "Will be filled after payload parsing"
        super().__init__(*args, **kwargs)
        self.reporter: Reporter = None  # type: ignore

        self.check_required_option("token", """
            The GitHub workflow token is not specified.

            Token is automatically created by GitHub Actions and stored in 
            ${{ secrets.GITHUB_TOKEN }} environment variable. For Universum
            to work correctly please copy it to ${{ env.GITHUB_TOKEN }}
            (as due to security reasons Universum cannot get access to secrets directly),
            or pass it directly using '--ghactions-token' ('-ght') command line parameter.
            """)

        self.payload: str = self.read_and_check_multiline_option("payload", """
            GitHub web-hook payload JSON is not specified.

            Please pass the payload stored in ${{ github.event_path }} file to this parameter
            directly via '--ghactions-payload' ('-ghp') command line parameter or by setting
            GITHUB_PAYLOAD environment variable, or by passing file path as the argument value
            (start filename with '@' character, e.g. '@/tmp/file.json' or '@payload.json' for
            relative path starting at current directory). Please note, that when passing
            a file, it's expected to be in UTF-8 encoding.
            """)

        try:
            self.payload_json: dict = json.loads(self.payload)
            self.settings.repo = self.payload_json['repository']['html_url']
            self.settings.refspec = self.payload_json['pull_request']['head']['ref']
        except json.decoder.JSONDecodeError as error:
            self.error(f"Provided payload value could not been parsed as JSON "
                       f"and returned the following error:\n {error}")

    def _clone(self, history_depth: int, destination_directory: str, clone_url: str) -> None:
        parsed_repo: urllib.parse.SplitResult = urllib.parse.urlsplit(clone_url)
        if parsed_repo.scheme == "https" and not parsed_repo.username:
            new_netloc = f"{self.settings.token}@{parsed_repo.netloc}"
            parsed_repo = (parsed_repo.scheme, new_netloc, parsed_repo.path,
                           parsed_repo.query, parsed_repo.fragment)  # type: ignore
        clone_url = str(urllib.parse.urlunsplit(parsed_repo))
        super()._clone(history_depth, destination_directory, clone_url)

    def code_review(self) -> Self:
        self.reporter = self.reporter_factory()
        self.reporter.subscribe(self)
        return self

    def update_review_version(self) -> None:
        self.out.log("GitHub has no review versions")

    def get_review_link(self) -> str:
        return self.payload_json['pull_request']['html_url']

    def is_latest_version(self) -> bool:
        return True

    def _report(self, url, request: dict) -> None:
        headers: Dict = {
            "Accept": "application/vnd.github+json",
            "Authorization": f"Bearer { self.settings.token }"
        }

        utils.make_request(url, request_method="POST", json=request, headers=headers, timeout=5*60)

    def code_report_to_review(self, report: dict) -> None:
        # git show returns string, each file separated by \n,
        # first line consists of commit id and commit comment, so it's skipped
        commit_files: List[str] = self.repo.git.show("--name-only", "--oneline",
                                                     self.settings.checkout_id).split('\n')[1:]
        # NB! When using GITHUB_TOKEN, the rate limit is 1,000 requests per hour per repository.
        # (https://docs.github.com/en/rest/overview/resources-in-the-rest-api?apiVersion=2022-11-28  ->
        #                                              #rate-limits-for-requests-from-github-actions)
        # Therefore the following reporting cycle will FAIL if PR has more than 1,000 analyzer issues
        for path, issues in report.items():
            if path not in commit_files:
                continue
            for issue in issues:
                request = dict(path=path,
                               commit_id=self.payload_json['pull_request']['head']['sha'],
                               body=issue['message'],
                               line=issue['line'],
                               side="RIGHT")
                self._report(self.payload_json['pull_request']['review_comments_url'], request)

    def report_start(self, report_text: str) -> None:
        pass

    def report_result(self, result: bool, report_text: Optional[str] = None, no_vote: bool = False):
        if not report_text:
            report_text = "Universum check finished"
        self._report(self.payload_json['pull_request']['comments_url'], {"body": report_text})



import json
from os import path

from .deployment_utils import UniversumRunner
from .utils import python

config = f"""
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Run script", artifacts="output.json",
                              command=["bash", "-c", "{python()} -m universum api file-diff > output.json"])])
"""


def test_error_wrong_environment(docker_main_and_nonci: UniversumRunner):
    cmd = f"{python()} -m universum api file-diff"
    log = docker_main_and_nonci.environment.assert_unsuccessful_execution(cmd)
    assert "Error: Failed to read the 'UNIVERSUM_DATA_FILE' from environment" in log


def test_p4_file_diff(docker_main_with_vcs: UniversumRunner):
    p4 = docker_main_with_vcs.perforce.p4
    p4_directory = docker_main_with_vcs.perforce.root_directory
    p4_file = docker_main_with_vcs.perforce.repo_file

    p4.run_edit(p4_file)
    p4_file.write_text("This line is added to the file.\n")
    p4.run_move(p4_file, p4_directory / "some_new_file_name.txt")
    change = p4.fetch_change()
    change["Description"] = "Rename basic config"
    shelve_cl = p4.save_change(change)[0].split()[1]
    p4.run_shelve("-fc", shelve_cl)

    log = docker_main_with_vcs.run(config, vcs_type="p4", environment=[f"SHELVE_CHANGELIST={shelve_cl}"],
                                   additional_parameters=" --p4-force-clean")
    assert "Module sh got exit code" not in log

    with open(path.join(docker_main_with_vcs.artifact_dir, "output.json"), encoding="utf-8") as f:
        result = json.load(f)

    assert result[0]["action"] == "move/add"
    assert result[0]["repo_path"] == "//depot/some_new_file_name.txt"
    assert result[1]["action"] == "move/delete"
    assert result[1]["repo_path"] == "//depot/writeable_file.txt"


def test_multiple_p4_file_diff(docker_main_with_vcs: UniversumRunner):
    p4 = docker_main_with_vcs.perforce.p4
    p4_directory = docker_main_with_vcs.perforce.root_directory

    for index in range(0, 10000):
        new_file = p4_directory / f"new_file_{index}.txt"
        new_file.write_text(f"This is file #{index}\n")
        p4.run_add(new_file)
    change = p4.fetch_change()
    change["Description"] = "Add 10000 files"
    shelve_cl = p4.save_change(change)[0].split()[1]
    p4.run_shelve("-fc", shelve_cl)

    log = docker_main_with_vcs.run(config, vcs_type="p4", environment=[f"SHELVE_CHANGELIST={shelve_cl}"],
                                   additional_parameters=" --p4-force-clean")
    assert "Module sh got exit code" not in log

    with open(path.join(docker_main_with_vcs.artifact_dir, "output.json"), encoding="utf-8") as f:
        result = json.load(f)
    assert len(result) == 10000
    for entry in result:
        assert entry["action"] == "add"


def test_git_file_diff(docker_main_with_vcs: UniversumRunner):
    repo = docker_main_with_vcs.git.repo
    server = docker_main_with_vcs.git.server
    logger = docker_main_with_vcs.git.logger
    git_directory = docker_main_with_vcs.git.root_directory
    git_file = docker_main_with_vcs.git.repo_file

    repo.git.checkout(server.target_branch)
    repo.git.checkout("new_testing_branch", b=True)
    repo.git.mv(git_file, git_directory / "some_new_file_name.txt")
    change = repo.index.commit("Special commit for testing")
    repo.remotes.origin.push(progress=logger, all=True)

    log = docker_main_with_vcs.run(config, vcs_type="git",
                                   environment=[f"GIT_CHERRYPICK_ID={change}",
                                                f"GIT_REFSPEC={server.target_branch}"])
    assert "Module sh got exit code" not in log

    with open(path.join(docker_main_with_vcs.artifact_dir, "output.json"), encoding="utf-8") as f:
        result = json.load(f)

    assert result[0]["action"] == "rename"
    assert result[0]["repo_path"] == "readme.txt"
    assert 'some_new_file_name.txt' in result[0]["local_path"]


def test_multiple_git_file_diff(docker_main_with_vcs: UniversumRunner):
    repo = docker_main_with_vcs.git.repo
    server = docker_main_with_vcs.git.server
    logger = docker_main_with_vcs.git.logger
    git_directory = docker_main_with_vcs.git.root_directory

    repo.git.checkout(server.target_branch)
    repo.git.checkout("new_testing_branch", b=True)
    files = []
    for index in range(0, 10000):
        new_file = git_directory / f"new_file_{index}.txt"
        new_file.write_text(f"This is file #{index}\n")
        files.append(str(new_file))
    repo.index.add(files)
    change = repo.index.commit("Special commit for testing")
    repo.remotes.origin.push(progress=logger, all=True)

    log = docker_main_with_vcs.run(config, vcs_type="git",
                                   environment=[f"GIT_CHERRYPICK_ID={change}",
                                                f"GIT_REFSPEC={server.target_branch}"])
    assert "Module sh got exit code" not in log

    with open(path.join(docker_main_with_vcs.artifact_dir, "output.json"), encoding="utf-8") as f:
        result = json.load(f)
    assert len(result) == 10000
    for entry in result:
        assert entry["action"] == "add"



import sys
from typing import ClassVar

from .base_output import BaseOutput
from .github_output import GithubOutput
from .html_output import HtmlOutput
from .teamcity_output import TeamcityOutput
from .terminal_based_output import TerminalBasedOutput
from ...lib import utils
from ...lib.gravity import Module, Dependency

__all__ = [
    "HasOutput",
    "MinimalOut",
    "Output"
]


class Output(Module):
    teamcity_driver_factory = Dependency(TeamcityOutput)
    terminal_driver_factory = Dependency(TerminalBasedOutput)
    html_driver_factory = Dependency(HtmlOutput)
    github_driver_factory = Dependency(GithubOutput)

    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Output", "Log appearance parameters")
        parser.add_argument("--out-type", "-ot", dest="type", choices=["tc", "term", "jenkins", "github"],
                            help="Type of output to produce (tc - TeamCity, jenkins - Jenkins, term - terminal, "
                                 "github - Github Actions). TeamCity, Jenkins and Github Actions environments are "
                                 "detected automatically when launched on build agent.")
        # `universum` -> html_log == default
        # `universum -hl` -> html_log == const
        # `universum -hl custom` -> html_log == custom
        parser.add_argument("--html-log", "-hl", nargs="?", const=HtmlOutput.default_name, default=None,
                            help=f"Generate a self-contained user-friendly HTML log. "
                                 f"Pass a desired log name as a parameter to this option, or a default "
                                 f"'{HtmlOutput.default_name}' will be used.")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.driver: BaseOutput = \
            utils.create_driver(local_factory=self.terminal_driver_factory,
                                teamcity_factory=self.teamcity_driver_factory,
                                jenkins_factory=self.terminal_driver_factory,
                                github_factory=self.github_driver_factory,
                                env_type=self.settings.type)
        self.html_driver = self._create_html_driver()

    def log_execution_start(self, title: str, version: str) -> None:
        self.driver.log_execution_start(title, version)
        self.html_driver.log_execution_start(title, version)

    def log_execution_finish(self, title: str, version: str) -> None:
        self.driver.log_execution_finish(title, version)
        self.html_driver.log_execution_finish(title, version)

    def log(self, line: str) -> None:
        self.driver.log(line)
        self.html_driver.log(line)

    def log_error(self, description: str) -> None:
        self.driver.log_error(description)
        self.html_driver.log_error(description)

    def log_external_command(self, command: str) -> None:
        self.driver.log_external_command(command)
        self.html_driver.log_external_command(command)

    def log_stdout(self, line: str) -> None:
        self.driver.log_stdout(line)
        self.html_driver.log_stdout(line)

    def log_stderr(self, line: str) -> None:
        self.driver.log_stderr(line)
        self.html_driver.log_stderr(line)

    def open_block(self, number: str, name: str) -> None:
        self.driver.open_block(number, name)
        self.html_driver.open_block(number, name)

    def close_block(self, number: str, name: str, status: str) -> None:
        self.driver.close_block(number, name, status)
        self.html_driver.close_block(number, name, status)

    def log_skipped(self, message: str) -> None:
        self.driver.log_skipped(message)
        self.html_driver.log_skipped(message)

    def log_summary_step(self, step_title: str, has_children: bool, status: str) -> None:
        self.driver.log_summary_step(step_title, has_children, status)
        self.html_driver.log_summary_step(step_title, has_children, status)

    # TODO: pass build problem to the Report module
    def report_build_problem(self, problem: str) -> None:
        self.driver.report_build_problem(problem)

    def set_build_status(self, status: str) -> None:
        self.driver.set_build_status(status)

    def _create_html_driver(self):
        is_enabled = self.settings.html_log is not None
        html_driver = self.html_driver_factory(log_name=self.settings.html_log) if is_enabled else None
        handler = HtmlDriverHandler(html_driver)
        return handler


class HasOutput(Module):
    out_factory: ClassVar = Dependency(Output)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.out: Output = self.out_factory()


class MinimalOut:
    def __init__(self, silent=False):
        self.silent = silent

    def log(self, line: str) -> None:
        if not self.silent:
            print(line)

    @staticmethod
    def report_build_problem(problem: str) -> None:
        pass

    @staticmethod
    def log_error(line: str) -> None:
        sys.stderr.write("Error: " + line)

    def log_execution_start(self, title, version):
        pass

    def log_execution_finish(self, title, version):
        pass


class HtmlDriverHandler:
    def __init__(self, driver):
        self.driver = driver

    def __getattr__(self, name):
        if self.driver:
            return getattr(self.driver, name)
        return lambda *args, **kwargs: None



# pylint: disable = redefined-outer-name

import os
from typing import Optional
import pytest

from universum.configuration_support import Configuration
from universum.modules.launcher import check_if_env_set


@pytest.fixture(autouse=True)
def os_environ():
    env_backup = dict(os.environ)
    yield
    os.environ = env_backup


def setup_env_vars(env_vars: Optional[dict]):
    if not env_vars:
        return

    for var, val in env_vars.items():
        os.environ[var] = val


def check(if_env_set_key: str, env_vars: Optional[dict]):
    setup_env_vars(env_vars)
    if_env_set_var = Configuration([dict(if_env_set=if_env_set_key)])[0]
    return check_if_env_set(if_env_set_var)


def assert_true(if_env_set_key: str, env_vars: Optional[dict] = None):
    assert check(if_env_set_key, env_vars) is True


def assert_false(if_env_set_key: str, env_vars: Optional[dict] = None):
    assert check(if_env_set_key, env_vars) is False


##########################################################################
# if_env_set var without operator
##########################################################################

def test_if_env_set_var_without_equal_operator_no_env_var():
    assert_false("VAR")


def test_if_env_set_var_without_equal_operator_env_var_set_to_true_1():
    assert_true("VAR", {"VAR": "Y"})


def test_if_env_set_var_without_equal_operator_env_var_set_to_true_2():
    assert_true("VAR_1 && VAR_2", {"VAR_1": "True", "VAR_2": "Yes"})


def test_if_env_set_var_without_equal_operator_env_var_not_set_to_true():
    assert_false("VAR", {"VAR": "False"})


def test_if_env_set_var_without_equal_operator_one_of_env_vars_not_set_to_true_1():
    assert_false("VAR_1 && VAR_2", {"VAR_1": "True", "VAR_2": "False"})


def test_if_env_set_var_without_equal_operator_one_of_env_vars_not_set_to_true_2():
    assert_false("VAR_1 && VAR_2", {"VAR_1": "False", "VAR_2": "True"})


##########################################################################
# Equal operator
##########################################################################

def test_if_env_set_equal_operator_no_env_var():
    assert_false("VAR == some &value")


def test_if_env_set_equal_operator_right_env_var():
    assert_true("VAR == some &value", {"VAR": "some &value"})


def test_if_env_set_equal_operator_wrong_env_var():
    assert_false("VAR == some &value", {"VAR": "some"})


def test_if_env_set_equal_operator_empty_string_no_env_var():
    assert_false("VAR == ")  # var == ""


def test_if_env_set_equal_operator_and_env_var_empty_strings():
    assert_true("VAR == ", {"VAR": ""})


def test_if_env_set_equal_operator_empty_string_env_var_is_value():
    assert_false("VAR == ", {"VAR": "True"})


##########################################################################
# Not equal operator
##########################################################################

def test_if_env_set_not_equal_operator_no_env_var():
    assert_true("VAR != True")


def test_if_env_set_not_equal_operator_right_env_var():
    assert_true("VAR != True", {"VAR": "value"})


def test_if_env_set_not_equal_operator_wrong_env_var():
    assert_false("VAR != True", {"VAR": "True"})


def test_if_env_set_not_equal_operator_empty_string_no_env_var():
    assert_true("VAR != ")  # var != ""


def test_if_env_set_not_equal_operator_and_env_var_empty_strings():
    assert_false("VAR != ", {"VAR": ""})


def test_if_env_set_not_equal_operator_empty_string_env_var_is_value():
    assert_true("VAR != ", {"VAR": "True"})


##########################################################################
# Other: var naming, spaces, quotes, empty if_env_set
##########################################################################

def test_if_env_set_is_empty():
    assert_true("")


def test_if_env_set_var_name_has_not_alphanumeric_characters():
    # if var name is not correct
    assert_false("VAR_&1 == value", {"VAR_&1": "value"})


def test_if_env_set_var_name_is_digit():
    assert_false("12 == value", {"12": "value"})


def test_if_env_set_check_if_hanging_spaces_removed():
    assert_true("  VAR   ==   =value  ", {"VAR": "=value"})


def test_if_env_set_check_config_with_no_spaces():
    assert_true("VAR===value", {"VAR": "=value"})


def test_if_env_set_check_if_spaces_removed_after_value():
    assert_false("VAR == value ", {"VAR": "value "})


def test_if_env_set_quotes_and_empty_env_var_value():
    assert_false("VAR == ''", {"VAR": ""})


def test_if_env_set_quotes_and_quotes_in_env_var_value():
    assert_true("VAR == ''", {"VAR": "''"})


##########################################################################
# if_env_set not in configuration
##########################################################################

def test_if_env_set_not_in_config():
    var = Configuration([dict(code_report=True),
                         dict(artifacts="htmlcov", command=["echo 123"], pass_tag="PASS")])
    os.environ["VAR_1"] = "False"
    os.environ["VAR_2"] = ""
    for item in var:
        assert check_if_env_set(item) is True


##########################################################################
# if_env_set for multiplication
##########################################################################

def assert_equal_multiplication(expected: list, env_vars: Optional[dict] = None):
    setup_env_vars(env_vars)
    var1 = Configuration([dict(if_env_set="VAR_1 == value_1"), dict(if_env_set="VAR_2 == value_2")])
    var2 = Configuration([dict(if_env_set=" && VAR_3 == value_3")])
    configs = var1 * var2
    result = list(filter(check_if_env_set, configs.all()))
    assert expected == result


def test_multiplication_filter_no_env_vars():
    assert_equal_multiplication([])


def test_multiplication_filter_one_env_var():
    assert_equal_multiplication([], {"VAR_1": "value_1"})


def test_multiplication_filter_one_step_true():
    assert_equal_multiplication([{'if_env_set': 'VAR_2 == value_2 && VAR_3 == value_3'}],
                                {"VAR_2": "value_2", "VAR_3": "value_3"})


def test_multiplication_filter_all_true():
    assert_equal_multiplication([{'if_env_set': 'VAR_1 == value_1 && VAR_3 == value_3'},
                                 {'if_env_set': 'VAR_2 == value_2 && VAR_3 == value_3'}],
                                {"VAR_1": "value_1", "VAR_2": "value_2", "VAR_3": "value_3"})



# pylint: disable = redefined-outer-name, abstract-method

import pathlib
import pytest

from universum.modules.vcs.github_app_vcs import GithubToken
from .conftest import FuzzyCallChecker
from .utils import BaseTestEnvironment, BaseVcsClient


class GithubHandlerEnvironment(BaseTestEnvironment):
    check_suite_payload: str = """
{
"action": "requested",
"repository": {
"url": "http://localhost"
},
"check_suite": {
"head_sha": "check_suite_head_sha"
},
"installation": {
"id": "installation_id"
}
}"""
    check_suite_url: str = "http://localhost/check-runs"
    check_run_payload: str = """
{
"action": "requested",
"check_run": {
"app": {
  "id": "INTEGRATION_ID"
},
"check_suite": {
    "head_branch": "check_run_check_suite_head_branch"
},
"head_sha": "check_run_head_sha",
"id": "check_run_id"
},
"repository": {
"clone_url": "repository_clone_url"
},
"installation": {
"id": "installation_id"
}
}"""
    check_run_url: str = "http://localhost/"

    def __init__(self, directory: pathlib.Path):
        super().__init__(BaseVcsClient(), directory, "github-handler", "")

        self.settings.GithubHandler.payload = "{}"
        self.settings.GithubHandler.event = "check_run"
        self.settings.GithubHandler.trigger_url = "http://localhost"

        self.settings.GithubToken.integration_id = "INTEGRATION_ID"
        self.settings.GithubToken.key = "this is key"


@pytest.fixture()
def github_handler_environment(tmp_path: pathlib.Path):
    yield GithubHandlerEnvironment(tmp_path)


@pytest.fixture(autouse=True)
def mock_token(monkeypatch: pytest.MonkeyPatch):
    def mocking_function(token_handler, parsed_id):
        assert parsed_id == 'installation_id'
        return "TOKEN_STRING"

    monkeypatch.setattr(GithubToken, 'get_token', mocking_function)


def test_success_github_handler_check_suite(github_handler_environment: GithubHandlerEnvironment):
    github_handler_environment.settings.GithubHandler.event = "check_suite"
    github_handler_environment.settings.GithubHandler.payload = github_handler_environment.check_suite_payload
    collected_http = github_handler_environment.run_with_http_server(url=github_handler_environment.check_suite_url,
                                                                     method="POST")
    collected_http.assert_request_headers_contained('Authorization', "token TOKEN_STRING")
    collected_http.assert_request_body_contained("head_sha", "check_suite_head_sha")


def test_success_github_handler_check_run(github_handler_environment: GithubHandlerEnvironment):
    github_handler_environment.settings.GithubHandler.event = "check_run"
    github_handler_environment.settings.GithubHandler.payload = github_handler_environment.check_run_payload
    collected_http = github_handler_environment.run_with_http_server(url=github_handler_environment.check_run_url)
    collected_http.assert_request_query_contained("GIT_REFSPEC", "check_run_check_suite_head_branch")
    collected_http.assert_request_query_contained("GIT_CHECKOUT_ID", "check_run_head_sha")
    collected_http.assert_request_query_contained("GITHUB_CHECK_ID", "check_run_id")
    collected_http.assert_request_query_contained("GIT_REPO", "repository_clone_url")
    collected_http.assert_request_query_contained("GITHUB_INSTALLATION_ID", "installation_id")


def test_error_github_handler_not_a_json(stdout_checker: FuzzyCallChecker,
                                         github_handler_environment: GithubHandlerEnvironment):
    github_handler_environment.settings.GithubHandler.payload = "not a JSON"
    github_handler_environment.run(expect_failure=True)
    stdout_checker.assert_has_calls_with_param("Provided payload value could not been parsed as JSON")


def test_error_github_handler_wrong_json_syntax(stdout_checker: FuzzyCallChecker,
                                                github_handler_environment: GithubHandlerEnvironment):
    github_handler_environment.settings.GithubHandler.payload = "{'key': 'value'}"
    github_handler_environment.run(expect_failure=True)
    stdout_checker.assert_has_calls_with_param("Provided payload value could not been parsed as JSON")


def test_error_github_handler_multiple_payloads(stdout_checker: FuzzyCallChecker,
                                                github_handler_environment: GithubHandlerEnvironment):
    github_handler_environment.settings.GithubHandler.payload = "[{},{}]"
    github_handler_environment.run(expect_failure=True)
    stdout_checker.assert_has_calls_with_param("Parsed payload JSON does not correspond to expected format")


def test_error_github_handler_empty_json(stdout_checker: FuzzyCallChecker,
                                         github_handler_environment: GithubHandlerEnvironment):
    github_handler_environment.settings.GithubHandler.payload = "{}"
    github_handler_environment.run(expect_failure=True)
    stdout_checker.assert_has_calls_with_param("Could not find key 'action' in provided payload")


def test_error_github_handler_json_missing_key(stdout_checker: FuzzyCallChecker,
                                               github_handler_environment: GithubHandlerEnvironment):
    github_handler_environment.settings.GithubHandler.event = "check_run"
    fixed_payload = github_handler_environment.check_run_payload.replace('"id": "installation_id"', '')
    github_handler_environment.settings.GithubHandler.payload = fixed_payload
    github_handler_environment.run(expect_failure=True)
    stdout_checker.assert_has_calls_with_param("Could not find key 'id' in provided payload")


def test_error_github_handler_wrong_event(stdout_checker: FuzzyCallChecker,
                                          github_handler_environment: GithubHandlerEnvironment):
    github_handler_environment.settings.GithubHandler.event = "unhandled_event"
    github_handler_environment.run()
    stdout_checker.assert_has_calls_with_param("Unhandled event, skipping...")


def test_error_github_handler_wrong_app(stdout_checker: FuzzyCallChecker,
                                        github_handler_environment: GithubHandlerEnvironment):
    github_handler_environment.settings.GithubHandler.event = "check_run"
    fixed_payload = github_handler_environment.check_run_payload.replace('"id": "INTEGRATION_ID"', '"id": "wrong_id"')
    github_handler_environment.settings.GithubHandler.payload = fixed_payload
    github_handler_environment.run()
    stdout_checker.assert_has_calls_with_param("Unhandled event, skipping...")


def test_error_github_handler_no_github_server(stdout_checker: FuzzyCallChecker,
                                               github_handler_environment: GithubHandlerEnvironment):
    github_handler_environment.settings.GithubHandler.event = "check_suite"
    github_handler_environment.settings.GithubHandler.payload = github_handler_environment.check_suite_payload
    github_handler_environment.run(expect_failure=True)
    stdout_checker.assert_has_calls_with_param("Failed to establish a new connection: [Errno 111] Connection refused")


def test_error_github_handler_no_jenkins_server(stdout_checker: FuzzyCallChecker,
                                                github_handler_environment: GithubHandlerEnvironment):
    github_handler_environment.settings.GithubHandler.event = "check_run"
    github_handler_environment.settings.GithubHandler.payload = github_handler_environment.check_run_payload
    github_handler_environment.run(expect_failure=True)
    stdout_checker.assert_has_calls_with_param("Failed to establish a new connection: [Errno 111] Connection refused")


def test_error_github_handler_wrong_schema(stdout_checker: FuzzyCallChecker,
                                           github_handler_environment: GithubHandlerEnvironment):
    github_handler_environment.settings.GithubHandler.event = "check_run"
    github_handler_environment.settings.GithubHandler.payload = github_handler_environment.check_run_payload
    github_handler_environment.settings.GithubHandler.trigger_url = "htttttp://localhost"
    github_handler_environment.run(expect_failure=True)
    stdout_checker.assert_has_calls_with_param("No connection adapters were found for 'htttttp://localhost'")


def test_error_github_handler_404(stdout_checker: FuzzyCallChecker, github_handler_environment: GithubHandlerEnvironment):
    github_handler_environment.settings.GithubHandler.event = "check_run"
    github_handler_environment.settings.GithubHandler.payload = github_handler_environment.check_run_payload
    url = github_handler_environment.check_run_url
    github_handler_environment.run_with_http_server(expect_failure=True, url=url, status='404')
    stdout_checker.assert_has_calls_with_param(f"404 Client Error: Not Found for url: {url}")



# pylint: disable = redefined-outer-name

import os
from typing import Callable, Union
import shutil
import pathlib
import pytest

from . import utils
from .conftest import FuzzyCallChecker
from .git_utils import GitTestEnvironment, GitClient
from .perforce_utils import P4TestEnvironment, PerforceWorkspace


@pytest.fixture()
def p4_submit_environment(perforce_workspace: PerforceWorkspace, tmp_path: pathlib.Path):
    yield P4TestEnvironment(perforce_workspace, tmp_path, test_type="submit")


@pytest.mark.parametrize("branch", ["write-protected", "trigger-protected"])
def test_p4_error_forbidden_branch(p4_submit_environment: P4TestEnvironment, branch: str):
    protected_dir = p4_submit_environment.vcs_client.root_directory / branch
    protected_dir.mkdir()
    file_name = utils.randomize_name("new_file") + ".txt"
    file_to_add = protected_dir / file_name
    text = "This is a new line in the file"
    file_to_add.write_text(text + "\n")

    p4_submit_environment.settings.Submit.reconcile_list = str(file_to_add)

    p4_submit_environment.run(expect_failure=True)

    p4 = p4_submit_environment.vcs_client.p4
    # make sure submitter didn't leave any pending CLs in the workspace
    assert not p4.run_changes("-c", p4_submit_environment.client_name, "-s", "pending")
    # make sure submitter didn't leave any pending changes in default CL
    assert not p4.run_opened("-C", p4_submit_environment.client_name)


def test_p4_success_files_in_default(p4_submit_environment: P4TestEnvironment):
    # This file should not be submitted, it should remain unchanged in default CL
    p4 = p4_submit_environment.vcs_client.p4
    p4_file = p4_submit_environment.vcs_client.repo_file
    p4.run_edit(str(p4_file))
    text = "This text should be in file"
    p4_file.write_text(text + "\n")

    # This file should be successfully submitted
    file_name = utils.randomize_name("new_file") + ".txt"
    new_file = p4_submit_environment.vcs_client.root_directory / file_name
    new_file.write_text("This is a new file" + "\n")

    p4_submit_environment.settings.Submit.reconcile_list = str(new_file)

    p4_submit_environment.run()
    assert text in p4_file.read_text()


def test_p4_error_files_in_default_and_reverted(p4_submit_environment: P4TestEnvironment):
    # This file should not be submitted, it should remain unchanged in default CL
    p4 = p4_submit_environment.vcs_client.p4
    p4_file = p4_submit_environment.vcs_client.repo_file
    p4.run_edit(str(p4_file))
    text_default = "This text should be in file"
    p4_file.write_text(text_default + "\n")

    # This file must fail submit and remain unchanged while not checked out any more
    protected_dir = p4_submit_environment.vcs_client.root_directory / "write-protected"
    protected_dir.mkdir()
    file_name = utils.randomize_name("new_file") + ".txt"
    new_file = protected_dir / file_name
    text_new = "This is a new line in the file"
    new_file.write_text(text_new + "\n")

    p4_submit_environment.settings.Submit.reconcile_list = str(new_file)

    p4_submit_environment.run(expect_failure=True)
    assert text_default in p4_file.read_text()
    assert text_new in new_file.read_text()


class SubmitterParameters:
    def __init__(self, stdout_checker: FuzzyCallChecker, environment: Union[GitTestEnvironment, P4TestEnvironment]):
        self.stdout_checker: FuzzyCallChecker = stdout_checker
        self.environment: Union[GitTestEnvironment, P4TestEnvironment] = environment

    def submit_path_list(self, path_list, expect_failure=False, **kwargs):
        self.environment.settings.Submit.reconcile_list = ",".join(path_list)

        if kwargs:
            for key, value in kwargs.items():
                setattr(self.environment.settings.Submit, key, value)

        self.environment.run(expect_failure=expect_failure)

    def assert_submit_success(self, path_list, **kwargs):
        self.submit_path_list(path_list, **kwargs)

        last_cl = self.environment.vcs_client.get_last_change()
        self.stdout_checker.assert_has_calls_with_param("==> Change " + last_cl + " submitted")

    def file_present(self, file_path):
        return self.environment.vcs_client.file_present(file_path)

    def text_in_file(self, text, file_path):
        return self.environment.vcs_client.text_in_file(text, file_path)


@pytest.fixture()
def submit_parameters(stdout_checker: FuzzyCallChecker):
    def inner(environment):
        return SubmitterParameters(stdout_checker, environment)
    yield inner


@pytest.fixture(params=["git", "p4"])
def submit_environment(request, perforce_workspace: PerforceWorkspace, git_client: GitClient, tmp_path: pathlib.Path):
    if request.param == "git":
        yield GitTestEnvironment(git_client, tmp_path, test_type="submit")
    else:
        yield P4TestEnvironment(perforce_workspace, tmp_path, test_type="submit")


def test_error_no_repo(submit_environment: Union[GitTestEnvironment, P4TestEnvironment], stdout_checker: FuzzyCallChecker):
    if submit_environment.settings.Vcs.type == "git":
        submit_environment.settings.ProjectDirectory.project_root = "non_existing_repo"
        submit_environment.run(expect_failure=True)
        stdout_checker.assert_has_calls_with_param("No such directory")
    else:
        submit_environment.settings.PerforceSubmitVcs.client = "non_existing_client"
        submit_environment.run(expect_failure=True)
        stdout_checker.assert_has_calls_with_param("Workspace 'non_existing_client' doesn't exist!")


def test_success_no_changes(submit_parameters: Callable, submit_environment: Union[GitTestEnvironment, P4TestEnvironment]):
    parameters = submit_parameters(submit_environment)
    parameters.submit_path_list([])


def test_success_commit_add_modify_remove_one_file(submit_parameters: Callable,
                                                   submit_environment: Union[GitTestEnvironment, P4TestEnvironment]):
    parameters = submit_parameters(submit_environment)

    file_name = utils.randomize_name("new_file") + ".txt"
    temp_file = parameters.environment.vcs_client.root_directory / file_name
    file_path = str(temp_file)

    # Add a file
    temp_file.write_text("This is a new file" + "\n")
    parameters.assert_submit_success([file_path])
    assert parameters.file_present(file_path)

    # Modify a file
    text = "This is a new line in the file"
    temp_file.write_text(text + "\n")
    parameters.assert_submit_success([file_path])
    assert parameters.text_in_file(text, file_path)

    # Delete a file
    temp_file.unlink()
    parameters.assert_submit_success([file_path])
    assert not parameters.file_present(file_path)


def test_success_ignore_new_and_deleted_while_edit_only(submit_parameters: Callable,
                                                        submit_environment: Union[GitTestEnvironment, P4TestEnvironment]):
    parameters = submit_parameters(submit_environment)

    new_file_name = utils.randomize_name("new_file") + ".txt"
    temp_file = parameters.environment.vcs_client.root_directory / new_file_name
    temp_file.write_text("This is a new temp file" + "\n")
    deleted_file_path = str(parameters.environment.vcs_client.repo_file)
    deleted_file_name = os.path.basename(deleted_file_path)
    os.remove(deleted_file_path)

    parameters.submit_path_list([str(temp_file), deleted_file_path], edit_only=True)

    parameters.stdout_checker.assert_has_calls_with_param(f"Skipping '{new_file_name}'")
    parameters.stdout_checker.assert_has_calls_with_param(f"Skipping '{deleted_file_name}'")
    parameters.stdout_checker.assert_has_calls_with_param("Nothing to submit")
    assert parameters.file_present(deleted_file_path)
    assert not parameters.file_present(str(temp_file))


def test_success_commit_modified_while_edit_only(submit_parameters: Callable,
                                                 submit_environment: Union[GitTestEnvironment, P4TestEnvironment]):
    parameters = submit_parameters(submit_environment)

    target_file = parameters.environment.vcs_client.repo_file
    text = utils.randomize_name("This is change ")
    target_file.write_text(text + "\n")

    parameters.assert_submit_success([str(target_file)], edit_only=True)
    assert parameters.text_in_file(text, str(target_file))


def test_error_review(submit_parameters: Callable, submit_environment: Union[GitTestEnvironment, P4TestEnvironment]):
    parameters = submit_parameters(submit_environment)

    target_file = parameters.environment.vcs_client.repo_file
    target_file.write_text("This is some change")

    parameters.submit_path_list([str(target_file)], review=True, expect_failure=True)
    parameters.stdout_checker.assert_has_calls_with_param("not supported")


def test_success_reconcile_directory(submit_parameters: Callable,
                                     submit_environment: Union[GitTestEnvironment, P4TestEnvironment]):
    parameters = submit_parameters(submit_environment)

    dir_name = utils.randomize_name("new_directory")

    # Create and reconcile new directory
    tmp_dir = parameters.environment.vcs_client.root_directory / dir_name
    tmp_dir.mkdir()
    for i in range(0, 9):
        tmp_file = tmp_dir / f"new_file{i}.txt"
        tmp_file.write_text("This is some file" + "\n")

    parameters.assert_submit_success([str(tmp_dir) + "/"])

    for i in range(0, 9):
        file_path = tmp_dir / f"new_file{i}.txt"
        assert parameters.file_present(str(file_path))

    # Create and reconcile a directory in a directory
    another_dir = tmp_dir / "another_directory"
    another_dir.mkdir()
    tmp_file = another_dir / "new_file.txt"
    tmp_file.write_text("This is some file" + "\n")

    parameters.assert_submit_success([str(tmp_dir) + "/"])
    assert parameters.file_present(str(tmp_file))

    # Modify some vcs
    text = utils.randomize_name("This is change ")
    for i in range(0, 9, 2):
        tmp_file = tmp_dir / f"new_file{i}.txt"
        tmp_file.write_text(text + "\n")

    parameters.assert_submit_success([str(tmp_dir) + "/"], edit_only=True)

    for i in range(0, 9, 2):
        file_path = tmp_dir / f"new_file{i}.txt"
        assert parameters.text_in_file(text, str(file_path))
    parameters.environment.settings.Submit.edit_only = False

    # Delete a directory
    shutil.rmtree(tmp_dir)
    parameters.assert_submit_success([str(tmp_dir)])
    assert not parameters.file_present(str(tmp_dir))


def test_success_reconcile_wildcard(submit_parameters: Callable,
                                    submit_environment: Union[GitTestEnvironment, P4TestEnvironment]):
    parameters = submit_parameters(submit_environment)

    dir_name = utils.randomize_name("new_directory")

    # Create embedded directories, partially reconcile
    tmp_dir = parameters.environment.vcs_client.root_directory / dir_name
    tmp_dir.mkdir()
    inner_dir = tmp_dir / "inner_directory"
    inner_dir.mkdir()
    text = "This is some file" + "\n"
    for i in range(0, 9):
        tmp_file = tmp_dir / f"new_file{i}.txt"
        tmp_file.write_text(text)
        tmp_file = tmp_dir / f"another_file{i}.txt"
        tmp_file.write_text(text)
        tmp_file = inner_dir / f"new_file{i}.txt"
        tmp_file.write_text(text)

    parameters.assert_submit_success([str(tmp_dir) + "/new_file*.txt"])

    for i in range(0, 9):
        file_name = f"new_file{i}.txt"
        file_path = tmp_dir / file_name
        assert parameters.file_present(str(file_path))
        file_path = inner_dir / file_name
        assert not parameters.file_present(str(file_path))
        file_name = f"another_file{i}.txt"
        file_path = tmp_dir / file_name
        assert not parameters.file_present(str(file_path))

    # Create one more directory
    other_dir_name = utils.randomize_name("new_directory")
    other_tmp_dir = parameters.environment.vcs_client.root_directory / other_dir_name
    other_tmp_dir.mkdir()
    for i in range(0, 9):
        tmp_file = other_tmp_dir / f"new_file{i}.txt"
        tmp_file.write_text("This is some file" + "\n")

    parameters.assert_submit_success([str(parameters.environment.vcs_client.root_directory) + "/new_directory*/"])

    for i in range(0, 9):
        file_name = f"new_file{i}.txt"
        file_path = other_tmp_dir / file_name
        assert parameters.file_present(str(file_path))
        file_path = inner_dir / file_name
        assert parameters.file_present(str(file_path))
        file_name = f"another_file{i}.txt"
        file_path = tmp_dir / file_name
        assert parameters.file_present(str(file_path))

    # Modify some vcs
    text = utils.randomize_name("This is change ")
    for i in range(0, 9, 2):
        tmp_file = tmp_dir / f"new_file{i}.txt"
        tmp_file.write_text(text + "\n")
        tmp_file = inner_dir / f"new_file{i}.txt"
        tmp_file.write_text(text + "\n")
        tmp_file = tmp_dir / f"another_file{i}.txt"
        tmp_file.write_text(text + "\n")

    parameters.assert_submit_success([str(tmp_dir) + "/new_file*.txt"], edit_only=True)

    for i in range(0, 9, 2):
        file_path = tmp_dir / f"new_file{i}.txt"
        assert parameters.text_in_file(text, str(file_path))
        file_path = inner_dir / f"new_file{i}.txt"
        assert not parameters.text_in_file(text, str(file_path))
        file_path = tmp_dir / f"another_file{i}.txt"
        assert not parameters.text_in_file(text, str(file_path))

    # Test subdirectory wildcard
    text = utils.randomize_name("This is change ")
    for i in range(1, 9, 2):
        tmp_file = tmp_dir / f"new_file{i}.txt"
        tmp_file.write_text(text + "\n")
        tmp_file = inner_dir / f"new_file{i}.txt"
        tmp_file.write_text(text + "\n")
        tmp_file = tmp_dir / f"another_file{i}.txt"
        tmp_file.write_text(text + "\n")

    parameters.assert_submit_success([str(tmp_dir) + "/*/*.txt"])

    for i in range(1, 9, 2):
        file_path = inner_dir / f"new_file{i}.txt"
        assert parameters.text_in_file(text, str(file_path))
        file_path = tmp_dir / f"new_file{i}.txt"
        assert not parameters.text_in_file(text, str(file_path))
        file_path = tmp_dir / f"another_file{i}.txt"
        assert not parameters.text_in_file(text, str(file_path))

    # Test edit-only subdirectory wildcard
    text = utils.randomize_name("This is change ")
    for i in range(0, 9, 3):
        tmp_file = tmp_dir / f"new_file{i}.txt"
        tmp_file.write_text(text + "\n")
        tmp_file = inner_dir / f"new_file{i}.txt"
        tmp_file.write_text(text + "\n")
        tmp_file = tmp_dir / "another_file{i}.txt"
        tmp_file.write_text(text + "\n")

        parameters.assert_submit_success([str(tmp_dir) + "/*/*.txt"], edit_only=True)

    for i in range(0, 9, 3):
        file_path = inner_dir / f"new_file{i}.txt"
        assert parameters.text_in_file(text, str(file_path))
        file_path = tmp_dir / f"new_file{i}.txt"
        assert not parameters.text_in_file(text, str(file_path))
        file_path = tmp_dir / f"another_file{i}.txt"
        assert not parameters.text_in_file(text, str(file_path))
    parameters.environment.settings.Submit.edit_only = False

    # Clean up the repo
    shutil.rmtree(str(tmp_dir))
    shutil.rmtree(str(other_tmp_dir))
    parameters.assert_submit_success([str(parameters.environment.vcs_client.root_directory) + "/*"])
    assert not parameters.file_present(str(tmp_dir))
    assert not parameters.file_present(str(other_tmp_dir))


def test_success_reconcile_partial(submit_parameters: Callable,
                                   submit_environment: Union[GitTestEnvironment, P4TestEnvironment]):
    # This test was failed when a bug in univesrum.lib.utils.unify_argument_list left empty entries in processed lists
    # When reconciling "", p4 adds to CL all changes made in scope of workspace (and therefore partial reconcile fails)

    parameters = submit_parameters(submit_environment)
    dir_name = utils.randomize_name("new_directory")
    tmp_dir = parameters.environment.vcs_client.root_directory / dir_name
    tmp_dir.mkdir()
    for i in range(0, 9):
        tmp_file = tmp_dir / f"new_file{i}.txt"
        tmp_file.write_text("This is some file" + "\n")

    reconcile_list = [str(tmp_dir / f"new_file{i}.txt") for i in range(0, 4)]
    reconcile_list.extend(["", " ", "\n"])
    parameters.assert_submit_success(reconcile_list)

    for i in range(0, 4):
        file_path = tmp_dir / f"new_file{i}.txt"
        assert parameters.file_present(str(file_path))

    for i in range(5, 9):
        file_path = tmp_dir / f"new_file{i}.txt"
        assert not parameters.file_present(str(file_path))

    # Delete a directory
    shutil.rmtree(tmp_dir)
    parameters.assert_submit_success([str(tmp_dir)])
    assert not parameters.file_present(str(tmp_dir))



# pylint: disable = redefined-outer-name

import pathlib
import pytest
import P4

from . import utils
from .conftest import FuzzyCallChecker
from .deployment_utils import UniversumRunner
from .perforce_utils import P4TestEnvironment, PerforceWorkspace
from .utils import python, LocalTestEnvironment


def test_which_universum_is_tested(docker_main: UniversumRunner, pytestconfig):
    # THIS TEST PATCHES ACTUAL SOURCES! Discretion is advised
    init_file = pytestconfig.rootpath / "universum" / "__init__.py"
    backup = init_file.read_bytes()
    test_line = utils.randomize_name("THIS IS A TESTING VERSION")
    init_file.write_text(f"""__title__ = "Universum"
__version__ = "{test_line}"
""")
    output = docker_main.run(utils.simple_test_config, vcs_type="none")
    init_file.write_bytes(backup)
    assert test_line in output

    docker_main.environment.assert_successful_execution("pip uninstall -y universum")
    docker_main.run(utils.simple_test_config, vcs_type="none", force_installed=True, expected_to_fail=True)
    docker_main.clean_artifacts()
    docker_main.run(utils.simple_test_config, vcs_type="none")  # not expected to fail
    if utils.reuse_docker_containers():
        docker_main.environment.install_python_module(docker_main.working_dir)


@pytest.fixture(name='print_text_on_teardown')
def fixture_print_text_on_teardown():
    yield
    print("TearDown fixture output must be handled by 'detect_fails' fixture")


def test_teardown_fixture_output_verification(print_text_on_teardown: None):
    pass


@pytest.mark.parametrize("should_not_execute", [True, False], ids=['no-sources', 'deleted-sources'])
def test_clean_sources_exception(tmp_path: pathlib.Path, stdout_checker: FuzzyCallChecker, should_not_execute):
    env = LocalTestEnvironment(tmp_path, "main")
    env.settings.Vcs.type = "none"
    source_directory = tmp_path
    if should_not_execute:
        source_directory = source_directory / 'nonexisting_dir'
    env.settings.LocalMainVcs.source_dir = str(source_directory)
    env.configs_file.write_text(f"""
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Test configuration",
                              command=["bash", "-c", "rm -rf {env.settings.ProjectDirectory.project_root}"])])
""")

    env.run(expect_failure=should_not_execute)
    error_message = f"[Errno 2] No such file or directory: '{env.settings.ProjectDirectory.project_root}'"
    stdout_checker.assert_has_calls_with_param(error_message)


def test_non_utf8_environment(docker_main: UniversumRunner):
    # POSIX has no 'UTF-8' in it's name, but supports Unicode
    output = docker_main.run("""
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Test configuration", command=["ls", "-la"])])
""", vcs_type="none", environment=['LANG=POSIX', 'LC_ALL=POSIX'])
    assert "\u2514" in output

    # 'en_US', unlike 'en_US.UTF-8', is latin-1
    docker_main.clean_artifacts()
    docker_main.environment.assert_successful_execution('apt install -y locales')
    docker_main.environment.assert_successful_execution('locale-gen --purge en_US')
    docker_main.environment.assert_successful_execution('update-locale LANG=en_US')
    output = docker_main.run("""
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Test configuration", command=["ls", "-la"])])
""", vcs_type="none", environment=['LANG=en_US', 'LC_ALL=en_US'])
    assert "\u2514" not in output


@pytest.fixture()
def perforce_environment(perforce_workspace: PerforceWorkspace, tmp_path: pathlib.Path):
    yield P4TestEnvironment(perforce_workspace, tmp_path, test_type="main")


def test_p4_multiple_spaces_in_mappings(perforce_environment: P4TestEnvironment):
    perforce_environment.settings.PerforceWithMappings.project_depot_path = None
    perforce_environment.settings.PerforceWithMappings.mappings = [f"{perforce_environment.vcs_client.depot}   /..."]
    perforce_environment.run()


def test_p4_repository_difference_format(perforce_environment: P4TestEnvironment):
    config = """
from universum.configuration_support import Configuration

configs = Configuration([dict(name="This is a changed step name", command=["ls", "-la"])])
"""
    perforce_environment.shelve_config(config)
    perforce_environment.run()
    diff = (perforce_environment.artifact_dir / 'REPOSITORY_DIFFERENCE.txt').read_text()
    assert "This is a changed step name" in diff
    assert "b'" not in diff


@pytest.fixture()
def mock_opened(monkeypatch):
    def mocking_function(*args, **kwargs):
        raise P4.P4Exception("Client 'p4_disposable_workspace' unknown - use 'client' command to create it.")

    monkeypatch.setattr(P4.P4, 'run_opened', mocking_function, raising=False)


@utils.nox_only
def test_p4_failed_opened(perforce_environment: P4TestEnvironment, mock_opened: None):
    perforce_environment.run()


# TODO: move this test to 'test_api.py' after test refactoring and Docker use reduction
@utils.nox_only
def test_p4_api_failed_opened(perforce_environment: P4TestEnvironment, mock_opened: None):
    step_name = "API"
    config = f"""
from universum.configuration_support import Step, Configuration

configs = Configuration([Step(name="{step_name}", artifacts="output.json",
                              command=["bash", "-c", "{python()} -m universum api file-diff > output.json"])])
    """
    perforce_environment.shelve_config(config)
    perforce_environment.settings.Launcher.output = "file"

    perforce_environment.run()
    log = (perforce_environment.artifact_dir / f'{step_name}_log.txt').read_text()
    assert "Module sh got exit code 1" in log
    assert "Getting file diff failed due to Perforce server internal error" in log


def test_p4_clean_empty_cl(perforce_environment: P4TestEnvironment, stdout_checker: FuzzyCallChecker):
    # This test creates an empty CL, triggering "file(s) not opened on this client" exception on cleanup
    # Wrong exception handling prevented further client cleanup on force clean, making final client deleting impossible

    config = f"""
from universum.configuration_support import Step, Configuration

configs = Configuration([Step(name="Create empty CL",
                              command=["bash", "-c",
                              "p4 --field 'Description=My pending change' --field 'Files=' change -o | p4 change -i"],
                              environment = {{"P4CLIENT": "{perforce_environment.client_name}",
                                              "P4PORT": "{perforce_environment.vcs_client.p4.port}",
                                              "P4USER": "{perforce_environment.vcs_client.p4.user}",
                                              "P4PASSWD": "{perforce_environment.vcs_client.p4.password}"}})])
"""
    perforce_environment.shelve_config(config)
    perforce_environment.run()
    error_message = f"""[Error]: "Client '{perforce_environment.client_name}' has pending changes."""
    stdout_checker.assert_absent_calls_with_param(error_message)


@pytest.fixture()
def perforce_environment_with_files(perforce_environment: P4TestEnvironment):
    files = [perforce_environment.vcs_client.create_file(utils.randomize_name("new_file") + ".txt")
             for _ in range(2)]

    yield {"env": perforce_environment, "files": files}

    for entry in files:
        perforce_environment.vcs_client.delete_file(entry.name)


def test_success_p4_resolve_unshelved(perforce_environment_with_files: dict, stdout_checker: FuzzyCallChecker):
    p4_file = perforce_environment_with_files["files"][0]
    env = perforce_environment_with_files["env"]
    config = f"""
from universum.configuration_support import Step, Configuration

configs = Configuration([Step(name="Print file", command=["bash", "-c", "cat '{p4_file.name}'"])])
"""
    env.shelve_config(config)
    cls = [env.vcs_client.shelve_file(p4_file, "This is changed line 1\nThis is unchanged line 2"),
           env.vcs_client.shelve_file(p4_file, "This is unchanged line 1\nThis is changed line 2")]
    env.settings.PerforceMainVcs.shelve_cls.extend(cls)

    env.run()
    stdout_checker.assert_has_calls_with_param("This is changed line 1")
    stdout_checker.assert_has_calls_with_param("This is changed line 2")


def test_fail_p4_resolve_unshelved(perforce_environment_with_files: dict, stdout_checker: FuzzyCallChecker):
    p4_file = perforce_environment_with_files["files"][0]
    env = perforce_environment_with_files["env"]
    config = f"""
from universum.configuration_support import Step, Configuration

configs = Configuration([Step(name="Print file", command=["bash", "-c", "cat '{p4_file.name}'"])])
"""
    env.shelve_config(config)
    cls = [env.vcs_client.shelve_file(p4_file, "This is changed line 1\nThis is unchanged line 2"),
           env.vcs_client.shelve_file(p4_file, "This is a different line 1\nThis is changed line 2")]
    env.settings.PerforceMainVcs.shelve_cls.extend(cls)

    env.run(expect_failure=True)
    stdout_checker.assert_has_calls_with_param("Problems during merge while resolving shelved CLs")
    stdout_checker.assert_has_calls_with_param(str(p4_file.name))


def test_success_p4_resolve_unshelved_multiple(perforce_environment_with_files: dict):
    p4_files = perforce_environment_with_files["files"]
    env = perforce_environment_with_files["env"]
    config = """
from universum.configuration_support import Step, Configuration

configs = Configuration([Step(name="Step one", command=["ls", "-l"])])
"""
    env.shelve_config(config)
    cl_1 = env.vcs_client.shelve_file(p4_files[0], "This is changed line 1\nThis is unchanged line 2")
    env.vcs_client.shelve_file(p4_files[1], "This is changed line 1\nThis is unchanged line 2", shelve_cl=cl_1)
    cl_2 = env.vcs_client.shelve_file(p4_files[0], "This is unchanged line 1\nThis is changed line 2")
    env.vcs_client.shelve_file(p4_files[1], "This is unchanged line 1\nThis is changed line 2", shelve_cl=cl_2)
    env.settings.PerforceMainVcs.shelve_cls.extend([cl_1, cl_2])

    env.run()
    repo_state = (env.artifact_dir / 'REPOSITORY_STATE.txt').read_text()
    assert p4_files[0].name in repo_state
    assert p4_files[1].name in repo_state


def test_exit_code_failed_report(perforce_environment: P4TestEnvironment):
    """
    This test checks for previous bug where exceptions during result reporting led to exit code 0
    even when '--fail-unsuccessful' option was enabled (and some steps failed)
    """
    config = """
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Unsuccessful step", command=["exit", "1"])])
"""
    perforce_environment.shelve_config(config)
    perforce_environment.settings.MainVcs.report_to_review = True
    perforce_environment.settings.Swarm.server_url = "some_server"
    perforce_environment.settings.Swarm.review_id = "some_id"
    perforce_environment.settings.Swarm.change = perforce_environment.settings.PerforceMainVcs.shelve_cls[0]
    perforce_environment.settings.Main.fail_unsuccessful = True

    perforce_environment.run(expect_failure=True)



import os

import urllib3

from ..error_state import HasErrorState
from ..output import HasOutput
from ..reporter import ReportObserver, Reporter
from ...lib import utils
from ...lib.ci_exception import CiException
from ...lib.gravity import Dependency

urllib3.disable_warnings(urllib3.exceptions.InsecurePlatformWarning)  # type: ignore

__all__ = [
    "Swarm"
]


def get_version_from_link(link):
    # link structure: http://<server.url>/reviews/<review_id>/tests/pass/C9B93E26-90B4-43C0-1A44-B076A3F51EE3.v1/
    try:
        version = link.split('.')[-1].strip('v').rstrip('/')
        int(version)
        return version
    except ValueError:
        return None


class Swarm(ReportObserver, HasOutput, HasErrorState):
    """
    This class contains CI functions for interaction with Swarm via 'swarm_cli.py'
    """
    reporter_factory = Dependency(Reporter)

    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Swarm",
                                                     "Parameters for performing a test run for pre-commit review")
        parser.add_argument("--swarm-server-url", "-ssu", dest="server_url", metavar="SWARM_SERVER",
                            help="Swarm server URL; is used for additional interaction such as voting for the review")
        parser.add_argument("--swarm-review-id", "-sre", dest="review_id", metavar="REVIEW",
                            help="Swarm review number; is sent by Swarm triggering link as '{review}'")
        parser.add_argument("--swarm-change", "-sch", dest="change", metavar="SWARM_CHANGELIST",
                            help="Swarm change list to unshelve; is sent by Swarm triggering link as '{change}'")
        parser.add_argument("--swarm-pass-link", "-spl", dest="pass_link", metavar="PASS",
                            help="Swarm 'success' link; is sent by Swarm triggering link as '{pass}'")
        parser.add_argument("--swarm-fail-link", "-sfl", dest="fail_link", metavar="FAIL",
                            help="Swarm 'fail' link; is sent by Swarm triggering link as '{fail}'")
        parser.add_argument("--swarm-review-versn", "-srv", dest="review_version", metavar="REVIEW_VERSION",
                            help="Swarm review version; is sent by Swarm triggering link as '{version}'")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = None
        self.ticket = None
        self.review_version = None
        if self.settings.review_version:
            self.review_version = self.settings.review_version
        self.review_latest_version = None
        self.client_root = ""
        self.mappings_dict = {}

        self.check_required_option("server_url", """
            The URL of the Swarm server is not specified.

            The URL is needed for communicating with swarm code review system: getting
            review revision, posting comments, voting. Please specify the server URL by
            using '--swarm-server-url' ('-ssu') command line parameter or by setting
            SWARM_SERVER environment variable.
            """)
        self.check_required_option("review_id", """
            The Swarm review number is not specified.

            The review number is needed for communicating with swarm code review system:
            getting review revision, posting comments, voting. Please specify the number by
            using '--swarm-review-id' ('-sre') command line parameter or by setting REVIEW
            environment variable.

            In order to setup sending of the review number for pre-commit in Swarm, please
            use the '{review}' argument in Automated Tests field of the Project Settings.
            """)
        self.check_required_option("change", """
            The Swarm changelist for unshelving is not specified.

            The changelist is used for unshelving change before build and for determining
            review revision. Please specify the changelist by using '--swarm-change'
            ('-sch') command line parameter or by setting SWARM_CHANGELIST environment
            variable.

            In order to setup sending of the review number for pre-commit in Swarm, please
            use the '{change}' argument in Automated Tests field of the Project Settings.
            """)

        if getattr(self.settings, "change", None) and (" " in self.settings.change or "," in self.settings.change):
            self.error("""
                The Swarm changelist for unshelving is incorrect.
                
                The changelist parameter must only contain one number. Please specify the
                changelist by using '--swarm-change' ('-sch') command line parameter or by
                setting SWARM_CHANGELIST environment variable.
                """)

        self.reporter = self.reporter_factory()
        self.reporter.subscribe(self)

    def get_review_link(self):
        return self.settings.server_url + "/reviews/" + self.settings.review_id + "/"

    def update_review_version(self):
        if self.review_version and self.review_latest_version:
            return

        result = utils.make_request(self.settings.server_url + "/api/v2/reviews/" + str(self.settings.review_id),
                                    critical=False, data={"id": self.settings.review_id},
                                    auth=(self.user, self.ticket))
        try:
            versions = result.json()["review"]["versions"]
        except (KeyError, ValueError) as e:
            text = "Error parsing Swarm server response. Full response is the following:\n"
            text += result.text
            raise CiException(text) from e

        self.review_latest_version = str(len(versions))

        if self.review_version:
            return
        self.out.log("Review version was not provided; trying to calculate it from PASS/FAIL links...")

        if self.settings.pass_link:
            self.review_version = get_version_from_link(self.settings.pass_link)
            if self.review_version:
                return
        if self.settings.fail_link:
            self.review_version = get_version_from_link(self.settings.fail_link)
            if self.review_version:
                return
        self.out.log("PASS/FAIL links either missing or have unexpected format; "
                     "try to calculate version from shelve number...")

        for index, entry in enumerate(versions):
            if int(entry["change"]) == int(self.settings.change):
                self.review_version = str(index + 1)
                return
        try:
            last_cl = int(versions[int(self.review_latest_version) - 1]["archiveChange"])
            if last_cl == int(self.settings.change):
                self.review_version = self.review_latest_version
        except KeyError:
            pass

    def is_latest_version(self):
        self.update_review_version()

        if self.review_latest_version == self.review_version:
            return True

        text = f"Current review version is {self.review_version}, " + \
               f"while latest review version is already {self.review_latest_version}"
        self.out.log(text)
        return False

    def post_comment(self, text, filename=None, line=None, version=None, no_notification=False):
        request = {"body": text,
                   "topic": "reviews/" + str(self.settings.review_id)}
        if filename:
            request["context[file]"] = filename
            if line:
                request["context[rightLine]"] = line
            if version:
                request["context[version]"] = version
            if no_notification:
                request["silenceNotification"] = "true"

        utils.make_request(self.settings.server_url + "/api/v9/comments", request_method="POST", critical=False,
                           data=request, auth=(self.user, self.ticket))

    def vote_review(self, result, version=None):
        request = {}
        if result:
            request["vote[value]"] = "up"
        else:
            request["vote[value]"] = "down"
        if version:
            request["vote[version]"] = version

        utils.make_request(self.settings.server_url + "/api/v6/reviews/" + self.settings.review_id, critical=False,
                           request_method="PATCH", data=request, auth=(self.user, self.ticket))

    def report_start(self, report_text):
        self.update_review_version()
        if self.review_version:
            report_text += "\nStarted build for review revision #" + self.review_version
        self.post_comment(report_text)

    def code_report_to_review(self, report):
        for path, issues in report.items():
            abs_path = os.path.join(self.client_root, path)
            if abs_path in self.mappings_dict:
                for issue in issues:
                    try:
                        self.post_comment(issue['message'], filename=self.mappings_dict[abs_path],
                                          line=issue['line'], no_notification=True)
                    except CiException as e:
                        self.out.log_error(str(e))

    def report_result(self, result, report_text=None, no_vote=False):
        # Opening links, sent by Swarm
        # Does not require login to Swarm; changes "Automated Tests" icon
        # Should not be applied to non-latest revision
        if result:
            link = self.settings.pass_link
        else:
            link = self.settings.fail_link

        if self.is_latest_version():
            if link is not None:
                self.out.log("Build status on Swarm will be updated via URL " + link)
                utils.make_request(link, critical=False)
            else:
                self.out.log("Build status on Swarm will not be updated because " +
                             f"the '{'PASS' if result else 'FAIL'}' link has not been provided.")
        else:
            text = "Build status on Swarm will not be updated because tested review revision is not latest."
            if not link:
                text += " Also, even if the review revision was latest, we wouldn't be able to report the status " + \
                        f"because the '{'PASS' if result else 'FAIL'}' link has not been provided."
            self.out.log(text)

        # Voting up or down; posting comments if any
        # An addition to "Automated Tests" functionality, requires login to Swarm
        self.update_review_version()
        if not no_vote:
            self.vote_review(result, version=self.review_version)
        if report_text:
            if self.review_version:
                report_text = "This is a build result for review revision #" + self.review_version + "\n" + report_text
            self.post_comment(report_text)
